# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2023 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Valerio Cosentino <valcos@bitergia.com>
#
import logging
import unittest
import time

import requests

from base import TestBaseBackend
from grimoire_elk.enriched.gitlab import NO_MILESTONE_TAG
from grimoire_elk.enriched.utils import REPO_LABELS
from grimoire_elk.raw.gitlab import GitLabOcean


class TestGitLab(TestBaseBackend):
    """Test GitLab backend"""

    connector = "gitlab"
    ocean_index = "test_" + connector
    enrich_index = "test_" + connector + "_enrich"
    ocean_index_anonymized = "test_" + connector + "_anonymized"
    enrich_index_anonymized = "test_" + connector + "_enrich_anonymized"

    def test_has_identites(self):
        """Test value of has_identities method"""

        enrich_backend = self.connectors[self.connector][2]()
        self.assertTrue(enrich_backend.has_identities())

    def test_items_to_raw(self):
        """Test whether JSON items are properly inserted into ES"""

        result = self._test_items_to_raw()

        self.assertGreater(result['items'], 0)
        self.assertGreater(result['raw'], 0)
        self.assertEqual(result['items'], result['raw'])

    def test_raw_to_enrich(self):
        """Test whether the raw index is properly enriched"""

        result = self._test_raw_to_enrich()

        self.assertGreater(result['raw'], 0)
        self.assertGreater(result['enrich'], 0)
        self.assertEqual(result['raw'], result['enrich'])

        enrich_backend = self.connectors[self.connector][2]()

        item = self.items[0]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(eitem['milestone'], "8.17")
        self.assertEqual(eitem['milestone_start_date'], "2017-01-07T00:00:00")
        self.assertEqual(eitem['milestone_due_date'], "2017-02-21T00:00:00")
        self.assertEqual(eitem['milestone_url'], "https://gitlab.com/gitlab-org/gitlab-ce/milestones/34")
        self.assertEqual(eitem['milestone_id'], 134231)
        self.assertEqual(eitem['milestone_iid'], 34)
        self.assertEqual(eitem['labels'], [])
        self.assertEqual(eitem['author_username'], "redfish64")
        self.assertEqual(eitem['author_uuid'], '3a9ab60b1586bf0a292ccb447746f7bcd1ab14e4')
        self.assertEqual(eitem['assignee_username'], 'redfish64')
        self.assertEqual(eitem['assignee_uuid'], '3a9ab60b1586bf0a292ccb447746f7bcd1ab14e4')
        self.assertEqual(eitem['repository'], "https://gitlab.com/fdroid/fdroiddata")

        item = self.items[1]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(eitem['milestone'], NO_MILESTONE_TAG)
        self.assertEqual(eitem['milestone_start_date'], None)
        self.assertEqual(eitem['milestone_due_date'], None)
        self.assertEqual(eitem['milestone_url'], None)
        self.assertEqual(eitem['milestone_id'], None)
        self.assertEqual(eitem['milestone_iid'], None)
        self.assertEqual(eitem['labels'], [])
        self.assertEqual(eitem['author_username'], "redfish64")
        self.assertEqual(eitem['author_uuid'], '3a9ab60b1586bf0a292ccb447746f7bcd1ab14e4')
        self.assertEqual(eitem['repository'], "https://gitlab.com/fdroid/fdroiddata")

        item = self.items[2]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(eitem['milestone'], NO_MILESTONE_TAG)
        self.assertEqual(eitem['milestone_start_date'], None)
        self.assertEqual(eitem['milestone_due_date'], None)
        self.assertEqual(eitem['milestone_url'], None)
        self.assertEqual(eitem['milestone_id'], None)
        self.assertEqual(eitem['milestone_iid'], None)
        self.assertEqual(eitem['labels'], ['CI/CD', 'Deliverable'])
        self.assertEqual(eitem['author_username'], "YoeriNijs")
        self.assertEqual(eitem['author_uuid'], '02a45079282ec8c5ee5e3d04cc5ef46441d45af8')
        self.assertEqual(eitem['repository'], "https://gitlab.com/fdroid/fdroiddata")

        item = self.items[4]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(eitem['milestone'], NO_MILESTONE_TAG)
        self.assertEqual(eitem['milestone_start_date'], None)
        self.assertEqual(eitem['milestone_due_date'], None)
        self.assertEqual(eitem['milestone_url'], None)
        self.assertEqual(eitem['milestone_id'], None)
        self.assertEqual(eitem['milestone_iid'], None)
        self.assertEqual(eitem['labels'], [])
        self.assertEqual(eitem['author_username'], None)
        self.assertEqual(eitem['repository'], "https://gitlab.com/gitlab-org/gitlab")

        item = self.items[5]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(eitem['milestone'], "8.17")
        self.assertEqual(eitem['milestone_start_date'], "2017-01-07T00:00:00")
        self.assertEqual(eitem['milestone_due_date'], "2017-02-21T00:00:00")
        self.assertEqual(eitem['milestone_url'], "https://gitlab.com/gitlab-org/gitlab-ce/milestones/34")
        self.assertEqual(eitem['milestone_id'], 134231)
        self.assertEqual(eitem['milestone_iid'], 34)
        self.assertEqual(eitem['labels'], [])
        self.assertEqual(eitem['author_username'], "grote")
        self.assertEqual(eitem['author_uuid'], 'c6f116d5e92c5ec1393542f63790c3c904ddba28')
        self.assertEqual(eitem['repository'], "https://gitlab.com/fdroid/fdroiddata")

        item = self.items[6]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(eitem['milestone'], NO_MILESTONE_TAG)
        self.assertEqual(eitem['milestone_start_date'], None)
        self.assertEqual(eitem['milestone_due_date'], None)
        self.assertEqual(eitem['milestone_url'], None)
        self.assertEqual(eitem['milestone_id'], None)
        self.assertEqual(eitem['milestone_iid'], None)
        self.assertEqual(eitem['labels'], [])
        self.assertEqual(eitem['author_username'], "Rudloff")
        self.assertEqual(eitem['author_uuid'], '56958a849936fca7a0e4361a6a08d99795770b2f')
        self.assertEqual(eitem['repository'], "https://gitlab.com/fdroid/fdroiddata")

        item = self.items[7]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(eitem['milestone'], NO_MILESTONE_TAG)
        self.assertEqual(eitem['milestone_start_date'], None)
        self.assertEqual(eitem['milestone_due_date'], None)
        self.assertEqual(eitem['milestone_url'], None)
        self.assertEqual(eitem['milestone_id'], None)
        self.assertEqual(eitem['milestone_iid'], None)
        self.assertEqual(eitem['labels'], ['CI/CD', 'Deliverable'])
        self.assertEqual(eitem['author_username'], "marc.nause")
        self.assertEqual(eitem['author_uuid'], 'b7bddb53fc3b7fbba0188015aa74bfc501a7f230')
        self.assertEqual(eitem['repository'], "https://gitlab.com/fdroid/fdroiddata")

    def test_enrich_repo_labels(self):
        """Test whether the field REPO_LABELS is present in the enriched items"""

        self._test_raw_to_enrich()
        enrich_backend = self.connectors[self.connector][2]()

        for item in self.items:
            eitem = enrich_backend.get_rich_item(item)
            self.assertIn(REPO_LABELS, eitem)

    def test_raw_to_enrich_sorting_hat(self):
        """Test enrich with SortingHat"""

        result = self._test_raw_to_enrich(sortinghat=True)
        self.assertGreater(result['raw'], 0)
        self.assertGreater(result['enrich'], 0)
        self.assertEqual(result['raw'], result['enrich'])

        enrich_backend = self.connectors[self.connector][2]()

        url = self.es_con + "/" + self.enrich_index + "/_search"
        response = enrich_backend.requests.get(url, verify=False).json()
        for hit in response['hits']['hits']:
            source = hit['_source']
            if 'author_uuid' in source:
                self.assertIn('author_domain', source)
                self.assertIn('author_gender', source)
                self.assertIn('author_gender_acc', source)
                self.assertIn('author_org_name', source)
                self.assertIn('author_bot', source)
                self.assertIn('author_multi_org_names', source)

    def test_raw_to_enrich_projects(self):
        """Test enrich with Projects"""

        result = self._test_raw_to_enrich(projects=True)
        # ... ?

    def test_refresh_identities(self):
        """Test refresh identities"""

        result = self._test_refresh_identities()
        # ... ?

    def test_items_to_raw_anonymized(self):
        """Test whether JSON items are properly inserted into ES anonymized"""

        result = self._test_items_to_raw_anonymized()

        self.assertGreater(result['items'], 0)
        self.assertGreater(result['raw'], 0)
        self.assertEqual(result['items'], result['raw'])

        item = self.items[0]['data']
        self.assertEqual(item['assignee']['username'], 'd04b492e0e54884033ddd4e32d4208a8340e4c85')
        self.assertEqual(item['assignee']['name'], '4e1e0c87bda6381324fb2b868859e1350c595c71')
        self.assertEqual(item['author']['username'], 'd04b492e0e54884033ddd4e32d4208a8340e4c85')
        self.assertEqual(item['author']['name'], '4e1e0c87bda6381324fb2b868859e1350c595c71')

        item = self.items[1]['data']
        self.assertIsNone(item['assignee'])
        self.assertEqual(item['author']['username'], 'd04b492e0e54884033ddd4e32d4208a8340e4c85')
        self.assertEqual(item['author']['name'], '4e1e0c87bda6381324fb2b868859e1350c595c71')

        item = self.items[4]['data']
        self.assertIsNone(item['assignee'])
        self.assertIsNone(item['author'])

        item = self.items[5]['data']
        self.assertEqual(item['merged_by']['username'], '5ef093e29634f927da333b7eaa479c248c93f26c')
        self.assertEqual(item['merged_by']['name'], '323866ff0cf9b5c7f817d1d444274582bef7d324')
        self.assertEqual(item['author']['username'], '499986947ba884c3c5946e15600a84f5fee8e9cb')
        self.assertEqual(item['author']['name'], 'ffcf43a6e72ca6d166ee38df916eee30d079d47c')

    def test_raw_to_enrich_anonymized(self):
        """Test whether the raw index is properly enriched"""

        result = self._test_raw_to_enrich_anonymized()

        self.assertGreater(result['raw'], 0)
        self.assertGreater(result['enrich'], 0)
        self.assertEqual(result['raw'], result['enrich'])

        enrich_backend = self.connectors[self.connector][2]()

        item = self.items[0]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(eitem['author_uuid'], 'b9f586e623a9e6e41df48f8ddcf097c0ca333b52')
        self.assertEqual(eitem['author_name'], '4e1e0c87bda6381324fb2b868859e1350c595c71')
        self.assertEqual(eitem['author_username'], 'd04b492e0e54884033ddd4e32d4208a8340e4c85')
        self.assertEqual(eitem['assignee_name'], '4e1e0c87bda6381324fb2b868859e1350c595c71')
        self.assertEqual(eitem['assignee_username'], 'd04b492e0e54884033ddd4e32d4208a8340e4c85')

        item = self.items[1]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(eitem['author_uuid'], 'b9f586e623a9e6e41df48f8ddcf097c0ca333b52')
        self.assertEqual(eitem['author_name'], '4e1e0c87bda6381324fb2b868859e1350c595c71')
        self.assertEqual(eitem['author_username'], 'd04b492e0e54884033ddd4e32d4208a8340e4c85')
        self.assertEqual(eitem['assignee_name'], 'Unknown')
        self.assertEqual(eitem['assignee_username'], None)

        item = self.items[4]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(eitem['author_uuid'], '')
        self.assertEqual(eitem['author_name'], 'Unknown')
        self.assertIsNone(eitem['author_username'])
        self.assertEqual(eitem['assignee_name'], 'Unknown')
        self.assertIsNone(eitem['assignee_username'])

        item = self.items[5]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(eitem['author_uuid'], '5a7534f25eaaeacf2d534172b6ed3f9effef94dd')
        self.assertEqual(eitem['author_name'], 'ffcf43a6e72ca6d166ee38df916eee30d079d47c')
        self.assertEqual(eitem['author_username'], '499986947ba884c3c5946e15600a84f5fee8e9cb')
        self.assertEqual(eitem['merge_author_name'], '323866ff0cf9b5c7f817d1d444274582bef7d324')
        self.assertEqual(eitem['merge_author_login'], '5ef093e29634f927da333b7eaa479c248c93f26c')

    def test_perceval_params(self):
        """Test the extraction of perceval params from an URL"""

        url = "https://gitlab.com/fdroid/fdroiddata"
        expected_params = [
            'fdroid', 'fdroiddata'
        ]
        self.assertListEqual(GitLabOcean.get_perceval_params_from_url(url), expected_params)

    def test_copy_raw_fields(self):
        """Test copied raw fields"""

        self._test_raw_to_enrich()
        enrich_backend = self.connectors[self.connector][2]()

        for item in self.items:
            eitem = enrich_backend.get_rich_item(item)
            for attribute in enrich_backend.RAW_FIELDS_COPY:
                if attribute in item:
                    self.assertEqual(item[attribute], eitem[attribute])
                else:
                    self.assertIsNone(eitem[attribute])

    def test_onion_study(self):
        """ Test that the onion study works correctly """

        alias = "all_onion"
        study, ocean_backend, enrich_backend = self._test_study('enrich_onion')
        study(ocean_backend, enrich_backend, alias, in_index='test_gitlab_enrich', out_index='test_gitlab_onion',
              data_source='gitlab-issues')

        url = self.es_con + "/_aliases"
        response = requests.get(url, verify=False).json()
        self.assertTrue('test_gitlab_onion' in response)

        time.sleep(1)

        url = self.es_con + "/test_gitlab_onion/_search?size=50"
        response = requests.get(url, verify=False).json()
        hits = response['hits']['hits']
        self.assertEqual(len(hits), 10)
        for hit in hits:
            source = hit['_source']
            self.assertIn('timeframe', source)
            self.assertIn('author_uuid', source)
            self.assertIn('author_name', source)
            self.assertIn('contributions', source)
            self.assertIn('metadata__timestamp', source)
            self.assertIn('project', source)
            self.assertIn('author_org_name', source)
            self.assertIn('cum_net_sum', source)
            self.assertIn('percent_cum_net_sum', source)
            self.assertIn('onion_role', source)
            self.assertIn('quarter', source)
            self.assertIn('metadata__enriched_on', source)
            self.assertIn('data_source', source)
            self.assertIn('grimoire_creation_date', source)


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, format='%(asctime)s %(message)s')
    logging.getLogger("urllib3").setLevel(logging.WARNING)
    logging.getLogger("requests").setLevel(logging.WARNING)
    unittest.main(warnings='ignore')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2023 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#   Georg Link <georglink@bitergia.com>
#   Quan`Zhou <quan@bitergia.com>
#

import json
import logging
import re
import sys

import pkg_resources
import requests
from elasticsearch import Elasticsearch, RequestsHttpConnection

from grimoirelab_toolkit.datetime import (datetime_to_utc,
                                          str_to_datetime,
                                          datetime_utcnow)
from perceval.backends.core.git import (GitCommand,
                                        GitRepository,
                                        EmptyRepositoryError,
                                        RepositoryError)
from .enrich import Enrich, metadata
from .study_ceres_aoc import areas_of_code, ESPandasConnector
from ..elastic_mapping import Mapping as BaseMapping
from ..elastic_items import HEADER_JSON, MAX_BULK_UPDATE_SIZE
from .utils import anonymize_url

GITHUB = 'https://github.com/'
DEMOGRAPHY_COMMIT_MIN_DATE = '1980-01-01'
logger = logging.getLogger(__name__)


class Mapping(BaseMapping):

    @staticmethod
    def get_elastic_mappings(es_major):
        """Get Elasticsearch mapping.

        :param es_major: major version of Elasticsearch, as string
        :returns:        dictionary with a key, 'items', with the mapping
        """

        mapping = """
        {
            "properties": {
               "message_analyzed": {
                    "type": "text",
                    "index": true
               }
           }
        }"""

        return {"items": mapping}


class GitEnrich(Enrich):

    mapping = Mapping

    # REGEX to extract authors from a multi author commit: several authors present
    # in the Author field in the commit. Used if self.pair_programming is True
    AUTHOR_P2P_REGEX = re.compile(r'(?P<first_authors>.* .*) ([aA][nN][dD]|&|\+) (?P<last_author>.* .*) (?P<email>.*)')
    AUTHOR_P2P_NEW_REGEX = re.compile(r"Co-authored-by:(?P<first_authors>.* .*)<(?P<email>.*)>\n?")

    # REGEX to extract authors from the commit
    AUTHOR_REGEX = re.compile(r'^\s*(?P<field>.*): (?P<author>.* <.*>)$')

    GIT_AOC_ENRICHED = "git_aoc-enriched"

    roles = ['Author', 'Commit']
    meta_fields = ['acked_by_multi', 'approved_by_multi', 'co_authored_by_multi', 'co_developed_by_multi',
                   'merged_by_multi', 'reported_by_multi', 'reviewed_by_multi', 'signed_off_by_multi',
                   'suggested_by_multi', 'tested_by_multi']
    meta_fields_suffixes = ['_bots', '_domains', '_names', '_org_names', '_uuids']
    meta_non_authored_prefix = 'non_authored_'

    def __init__(self, db_sortinghat=None, json_projects_map=None,
                 db_user='', db_password='', db_host='', db_path=None,
                 db_port=None, db_ssl=False, db_verify_ssl=True, db_tenant=None,
                 pair_programming=False):
        super().__init__(db_sortinghat=db_sortinghat, json_projects_map=json_projects_map,
                         db_user=db_user, db_password=db_password, db_host=db_host,
                         db_port=db_port, db_path=db_path, db_ssl=db_ssl, db_verify_ssl=db_verify_ssl,
                         db_tenant=db_tenant)

        self.studies = []
        self.studies.append(self.enrich_demography)
        self.studies.append(self.enrich_areas_of_code)
        self.studies.append(self.enrich_onion)
        self.studies.append(self.enrich_git_branches)
        self.studies.append(self.enrich_forecast_activity)
        self.studies.append(self.enrich_extra_data)

        self.rate_limit = None
        self.rate_limit_reset_ts = None
        self.min_rate_to_sleep = 100  # if pending rate < 100 sleep
        self.pair_programming = pair_programming

    def get_field_author(self):
        return "Author"

    def get_field_unique_id(self):
        # In pair_programming the uuid is not unique: a commit can create
        # several commits
        if self.pair_programming:
            return "git_uuid"
        return "uuid"

    def get_field_date(self):
        """ Field with the date in the JSON enriched items """
        return "grimoire_creation_date"

    def __get_authors(self, authors_str):
        # Extract the authors from a multiauthor

        m = self.AUTHOR_P2P_REGEX.match(authors_str)
        if m:
            authors = m.group('first_authors').split(",")
            authors = [author.strip() for author in authors]
            authors += [m.group('last_author')]

        n = self.AUTHOR_P2P_NEW_REGEX.findall(authors_str)
        if n:
            for i in n:
                authors += [i[0]]
        # Remove duplicates
        authors = list(set(authors))

        return authors

    def get_identities(self, item):
        """Return the identities from an item."""

        if item['data']['Author']:
            # Check multi authors commits
            m = self.AUTHOR_P2P_REGEX.match(item['data']["Author"])
            n = self.AUTHOR_P2P_NEW_REGEX.match(item['data']["Author"])
            if (m or n) and self.pair_programming:
                authors = self.__get_authors(item['data']["Author"])
                for author in authors:
                    user = self.get_sh_identity(author)
                    yield user
            else:
                user = self.get_sh_identity(item['data']["Author"])
                yield user
        if item['data']['Commit']:
            m = self.AUTHOR_P2P_REGEX.match(item['data']["Commit"])
            n = self.AUTHOR_P2P_NEW_REGEX.match(item['data']["Author"])
            if (m or n) and self.pair_programming:
                committers = self.__get_authors(item['data']['Commit'])
                for committer in committers:
                    user = self.get_sh_identity(committer)
                    yield user
            else:
                user = self.get_sh_identity(item['data']['Commit'])
                yield user
        if 'Signed-off-by' in item['data'] and self.pair_programming:
            signers = item['data']["Signed-off-by"]
            for signer in signers:
                user = self.get_sh_identity(signer)
                yield user

    def get_sh_identity(self, item, identity_field=None):
        # John Smith <john.smith@bitergia.com>
        identity = {}

        git_user = item  # by default a specific user dict is expected
        if isinstance(item, dict) and 'data' in item:
            git_user = item['data'][identity_field]

        fields = git_user.split("<")
        name = fields[0]
        name = name.strip()  # Remove space between user and email
        email = None
        if len(fields) > 1:
            email = git_user.split("<")[1][:-1]
        identity['username'] = None
        identity['email'] = email
        identity['name'] = name

        return identity

    def get_project_repository(self, eitem):
        return eitem['origin']

    @metadata
    def get_rich_item(self, item):

        eitem = {}
        self.copy_raw_fields(self.RAW_FIELDS_COPY, item, eitem)
        # For pair programming uuid is not a unique field. Use git_uuid in general as unique field.
        eitem['git_uuid'] = eitem['uuid']
        # The real data
        commit = item['data']

        # data fields to copy
        copy_fields = ["message"]
        for f in copy_fields:
            if f in commit:
                eitem[f] = commit[f]
            else:
                eitem[f] = None
        # Fields which names are translated
        map_fields = {"commit": "hash", "message": "message_analyzed"}
        for fn in map_fields:
            if fn in commit:
                eitem[map_fields[fn]] = commit[fn]
            else:
                eitem[map_fields[fn]] = None

        if 'message' in commit:
            eitem['message'] = commit['message'][:self.KEYWORD_MAX_LENGTH]

        if 'refs' in commit:
            eitem["commit_tags"] = list(filter(lambda r: "tag: " in r, commit['refs']))

        eitem['hash_short'] = eitem['hash'][0:6]
        # Enrich dates

        author_date = self.__cast_str_to_datetime(commit, 'AuthorDate')
        commit_date = self.__cast_str_to_datetime(commit, 'CommitDate')

        eitem["author_date"] = author_date.replace(tzinfo=None).isoformat()
        eitem["commit_date"] = commit_date.replace(tzinfo=None).isoformat()

        eitem["author_date_weekday"] = author_date.replace(tzinfo=None).isoweekday()
        eitem["author_date_hour"] = author_date.replace(tzinfo=None).hour

        eitem["commit_date_weekday"] = commit_date.replace(tzinfo=None).isoweekday()
        eitem["commit_date_hour"] = commit_date.replace(tzinfo=None).hour

        utc_author_date = datetime_to_utc(author_date)
        utc_commit_date = datetime_to_utc(commit_date)

        eitem["utc_author"] = utc_author_date.replace(tzinfo=None).isoformat()
        eitem["utc_commit"] = utc_commit_date.replace(tzinfo=None).isoformat()

        eitem["utc_author_date_weekday"] = utc_author_date.replace(tzinfo=None).isoweekday()
        eitem["utc_author_date_hour"] = utc_author_date.replace(tzinfo=None).hour

        eitem["utc_commit_date_weekday"] = utc_commit_date.replace(tzinfo=None).isoweekday()
        eitem["utc_commit_date_hour"] = utc_commit_date.replace(tzinfo=None).hour

        eitem["tz"] = int(author_date.strftime("%z")[0:3])
        eitem["branches"] = []

        # Compute time to commit
        time_to_commit_delta = datetime_to_utc(author_date) - datetime_to_utc(commit_date)
        eitem["time_to_commit_hours"] = round(time_to_commit_delta.seconds / 3600, 2)

        # Other enrichment
        eitem["repo_name"] = item["origin"]

        if eitem["repo_name"].startswith('http'):
            eitem["repo_name"] = anonymize_url(eitem["repo_name"])

        # Number of files touched
        eitem["files"] = 0
        # Number of lines added and removed
        lines_added = 0
        lines_removed = 0
        for cfile in commit["files"]:
            if 'action' not in cfile:
                # merges are not counted
                continue
            eitem["files"] += 1
            if 'added' in cfile and 'removed' in cfile:
                try:
                    lines_added += int(cfile["added"])
                    lines_removed += int(cfile["removed"])
                except ValueError:
                    # logger.warning(cfile)
                    continue
        eitem["lines_added"] = lines_added
        eitem["lines_removed"] = lines_removed
        eitem["lines_changed"] = lines_added + lines_removed

        # author_name and author_domain are added always
        identity = self.get_sh_identity(commit["Author"])
        eitem["author_name"] = identity['name']
        eitem["author_domain"] = self.get_identity_domain(identity)

        # committer data
        identity = self.get_sh_identity(commit["Commit"])
        eitem["committer_name"] = identity['name']
        eitem["committer_domain"] = self.get_identity_domain(identity)

        # title from first line
        if 'message' in commit:
            eitem["title"] = commit['message'].split('\n')[0]
        else:
            eitem["title"] = None

        # If it is a github repo, include just the repo string
        if GITHUB in item['origin']:
            eitem['github_repo'] = item['origin'].replace(GITHUB, '')
            eitem['github_repo'] = re.sub('.git$', '', eitem['github_repo'])
            eitem["url_id"] = eitem['github_repo'] + "/commit/" + eitem['hash']

        if 'project' in item:
            eitem['project'] = item['project']

        # Adding the git author domain
        author_domain = self.get_identity_domain(self.get_sh_identity(item, 'Author'))
        eitem['git_author_domain'] = author_domain

        grimoire_fields = self.get_grimoire_fields(author_date, "commit")
        eitem.update(grimoire_fields)

        # grimoire_creation_date is needed in the item
        item.update(grimoire_fields)
        eitem.update(self.get_item_sh(item, self.roles))

        if self.prjs_map:
            eitem.update(self.get_item_project(eitem))

        if self.pair_programming:
            eitem = self.__add_pair_programming_metrics(commit, eitem)

        self.add_repository_labels(eitem)
        self.add_metadata_filter_raw(eitem)
        self.__add_commit_meta_fields(eitem, commit)
        return eitem

    def __cast_str_to_datetime(self, item, attribute):
        """Convert str to datetime fixing possible errors"""

        field_date = str_to_datetime(item[attribute])

        try:
            _ = int(field_date.strftime("%z")[0:3])
        except ValueError:
            logger.warning("[git] {} in commit {} has a wrong format".format(
                           attribute, item['commit']))
            return field_date.replace(tzinfo=None).isoformat()
        return field_date

    def __add_commit_meta_fields(self, eitem, commit):
        """Add commit meta fields as signed_off_by, reviwed_by, tested_by, etc."""
        non_authored = []
        if self.meta_non_authored_prefix:
            non_authored = [self.meta_non_authored_prefix + field for field in self.meta_fields]
        all_meta_fields = self.meta_fields + non_authored

        for field in all_meta_fields:
            for suffix in self.meta_fields_suffixes:
                eitem[field + suffix] = []

        if 'message' not in commit:
            return

        meta_eitem = {}
        for line in commit['message'].split('\n'):
            m = self.AUTHOR_REGEX.match(line)
            if not m:
                continue

            meta_field = m.group('field').lower().replace('-', '_') + '_multi'
            if meta_field not in self.meta_fields:
                continue

            author = m.group('author')
            identity = self.get_sh_identity(author)

            if self.sortinghat:
                # Create SH identity if it does not exist
                backend_name = self.get_sh_backend_name()
                identity_id = self.generate_uuid(backend_name,
                                                 email=identity['email'],
                                                 name=identity['name'],
                                                 username=identity['username'])
                individual = self.get_entity(identity_id)
                if not individual:
                    identity_tuple = tuple(identity.items())
                    self.add_sh_identity_cache(identity_tuple)
                    logger.debug("Create a new individual {} in commit_meta_fields".format(identity_id))
                item_date = eitem[self.get_field_date()]
                sh_fields = self.get_item_sh_fields(identity, item_date, rol=meta_field)
            else:
                sh_fields = self.get_item_no_sh_fields(identity, rol=meta_field)

            uuid = sh_fields[meta_field + '_uuid']
            self.add_meta_fields(eitem, meta_eitem, sh_fields, meta_field, uuid, self.meta_fields_suffixes,
                                 self.meta_non_authored_prefix)
        eitem.update(meta_eitem)

    def __add_pair_programming_metrics(self, commit, eitem):

        def get_pair_programming_metrics(eitem, nauthors):
            ndecimals = 2
            metrics = {}
            files = eitem['files']
            ladded = eitem['lines_added']
            lremoved = eitem['lines_removed']
            lchanged = eitem['lines_changed']
            metrics['pair_programming_commit'] = round(1.0 / nauthors, ndecimals)
            metrics['pair_programming_files'] = round(files / nauthors, ndecimals)
            metrics["pair_programming_lines_added"] = round(ladded / nauthors, ndecimals)
            metrics["pair_programming_lines_removed"] = round(lremoved / nauthors, ndecimals)
            metrics["pair_programming_lines_changed"] = round(lchanged / nauthors, ndecimals)

            return metrics

        # Include pair programming metrics in all cases. In general, 1 author.
        eitem.update(get_pair_programming_metrics(eitem, 1))

        # Multi author support
        eitem['is_git_commit_multi_author'] = 0
        if 'is_git_commit_multi_author' in commit:
            eitem['is_git_commit_multi_author'] = commit['is_git_commit_multi_author']
        if 'authors' in commit:
            eitem['authors'] = commit['authors']
            nauthors = len(commit['authors'])
            eitem.update(get_pair_programming_metrics(eitem, nauthors))

        # Pair Programming support using Signed-off
        eitem['Signed-off-by_number'] = 0
        eitem['is_git_commit_signed_off'] = 0
        if 'Signed-off-by' in commit:
            eitem['Signed-off-by'] = commit['Signed-off-by']
            eitem['Signed-off-by_number'] = len(commit['Signed-off-by'])
            if 'is_git_commit_signed_off' in commit:
                # Commits generated for signed_off people
                eitem['is_git_commit_signed_off'] = commit['is_git_commit_signed_off']
            # The commit for the original Author also needs this data
            eitem['authors_signed_off'] = commit['authors_signed_off']
            nauthors = len(commit['authors_signed_off'])
            eitem.update(get_pair_programming_metrics(eitem, nauthors))
        return eitem

    def enrich_items(self, ocean_backend, events=False):
        """ Implementation supporting signed-off and multiauthor/committer commits.
        Multiauthor/Multcommiter commits are the ones authored/commited by more
        than one users, in such a case, ELK first extracts authors names from
        raw data through regex, for instance in,

        ...
        "data": {
            "Author": "Eduardo Morais and Zhongpeng Lin <companheiro.vermelho@gmail.com>",
            "AuthorDate": "Tue Aug 14 14:32:15 2012 -0300",
            "Commit": "Eduardo Morais and Zhongpeng Lin <companheiro.vermelho@gmail.com>",
            "CommitDate": "Tue Aug 14 14:32:15 2012 -0300",
            "Signed-off-by": [
                "Eduardo Morais <companheiro.vermelho@gmail.com>"
            ],
            "commit": "87783129c3f00d2c81a3a8e585eb86a47e39891a",
        ...

        authors extracted are ['Eduardo Morais', 'Zhongpeng Lin'] (can be more than 2).
        A new rich item is now created using each author name. For multicommitter only
        the first committer name is used to create the rich item. In case the commit is
        signed-off by committer, raw data has extra "Signed-off-by" attribute used to
        create a new rich item for every author who signed off the commit, making sure
        duplicate entries are not created.

        Note -
            "message": "Enable users to pass flags\n\nCo-authored-by: mariiapunda <mariiapunda@users.noreply.github.com>",
        Co-authored commits like these are not considered as multiauthored commits in ELK.
        """
        headers = {"Content-Type": "application/json"}

        max_items = self.elastic.max_items_bulk
        current = 0
        total = 0
        bulk_json = ""

        total_signed_off = 0
        total_multi_author = 0

        url = self.elastic.get_bulk_url()

        logger.debug("[git] Adding items to {} (in {} packs)".format(anonymize_url(url), max_items))
        items = ocean_backend.fetch()

        for item in items:
            if self.pair_programming:
                # First we need to add the authors field to all commits
                # Check multi author
                m = self.AUTHOR_P2P_REGEX.match(item['data']['Author'])
                n = self.AUTHOR_P2P_NEW_REGEX.match(item['data']['Author'])
                if m or n:
                    logger.debug("[git] Multiauthor detected. Creating one commit "
                                 "per author: {}".format(item['data']['Author']))
                    item['data']['authors'] = self.__get_authors(item['data']['Author'])
                    item['data']['Author'] = item['data']['authors'][0]
                    item['data']['is_git_commit_multi_author'] = 1
                m = self.AUTHOR_P2P_REGEX.match(item['data']['Commit'])
                n = self.AUTHOR_P2P_NEW_REGEX.match(item['data']['Author'])
                if m or n:
                    logger.debug("[git] Multicommitter detected: using just the first committer")
                    item['data']['committers'] = self.__get_authors(item['data']['Commit'])
                    item['data']['Commit'] = item['data']['committers'][0]
                # Add the authors list using the original Author and the Signed-off list
                if 'Signed-off-by' in item['data']:
                    authors_all = item['data']['Signed-off-by'] + [item['data']['Author']]
                    item['data']['authors_signed_off'] = list(set(authors_all))

            if current >= max_items:
                try:
                    total += self.elastic.safe_put_bulk(url, bulk_json)
                    json_size = sys.getsizeof(bulk_json) / (1024 * 1024)
                    logger.debug("[git] Added {} items to {} ({:.2f} MB)".format(
                                 total, anonymize_url(url), json_size))
                except UnicodeEncodeError:
                    # Why is requests encoding the POST data as ascii?
                    logger.warning("[git] Unicode error in enriched items, converting to ascii")
                    safe_json = str(bulk_json.encode('ascii', 'ignore'), 'ascii')
                    total += self.elastic.safe_put_bulk(url, safe_json)
                bulk_json = ""
                current = 0

            rich_item = self.get_rich_item(item)
            data_json = json.dumps(rich_item)
            unique_field = self.get_field_unique_id()
            bulk_json += '{"index" : {"_id" : "%s" } }\n' % (rich_item[unique_field])
            bulk_json += data_json + "\n"  # Bulk document
            current += 1

            if self.pair_programming:
                # Multi author support
                if 'authors' in item['data']:
                    # First author already added in the above commit
                    authors = item['data']['authors']
                    for i in range(1, len(authors)):
                        # logger.debug('Adding a new commit for %s', authors[i])
                        item['data']['Author'] = authors[i]
                        item['data']['is_git_commit_multi_author'] = 1
                        rich_item = self.get_rich_item(item)
                        commit_id = item["uuid"] + "_" + str(i - 1)
                        rich_item['git_uuid'] = commit_id
                        data_json = json.dumps(rich_item)
                        bulk_json += '{"index" : {"_id" : "%s" } }\n' % rich_item['git_uuid']
                        bulk_json += data_json + "\n"  # Bulk document
                        current += 1
                        total_multi_author += 1

                if rich_item['Signed-off-by_number'] > 0:
                    nsg = 0
                    # Remove duplicates and the already added Author if exists
                    authors = list(set(item['data']['Signed-off-by']))
                    if item['data']['Author'] in authors:
                        authors.remove(item['data']['Author'])
                    for author in authors:
                        # logger.debug('Adding a new commit for %s', author)
                        # Change the Author in the original commit and generate
                        # a new enriched item with it
                        item['data']['Author'] = author
                        item['data']['is_git_commit_signed_off'] = 1
                        rich_item = self.get_rich_item(item)
                        commit_id = item["uuid"] + "_" + str(nsg)
                        rich_item['git_uuid'] = commit_id
                        data_json = json.dumps(rich_item)
                        bulk_json += '{"index" : {"_id" : "%s" } }\n' % rich_item['git_uuid']
                        bulk_json += data_json + "\n"  # Bulk document
                        current += 1
                        total_signed_off += 1
                        nsg += 1

        if current > 0:
            total += self.elastic.safe_put_bulk(url, bulk_json)

        if total == 0:
            # No items enriched, nothing to upload to ES
            return total

        if self.pair_programming:
            logger.info("[git] Signed-off commits generated: {}".format(total_signed_off))
            logger.info("[git] Multi author commits generated: {}".format(total_multi_author))

        return total

    def enrich_demography(self, ocean_backend, enrich_backend, alias, date_field="grimoire_creation_date",
                          author_field="author_uuid"):

        super().enrich_demography(ocean_backend, enrich_backend, alias, date_field, author_field=author_field)

    def enrich_areas_of_code(self, ocean_backend, enrich_backend, alias, no_incremental=False,
                             in_index="git-raw",
                             out_index=GIT_AOC_ENRICHED,
                             sort_on_field='metadata__timestamp'):

        log_prefix = "[git] study areas_of_code"

        logger.info("{} Starting study - Input: {} Output: {}".format(log_prefix, in_index, out_index))

        # Creating connections
        es_in = Elasticsearch([ocean_backend.elastic.url], retry_on_timeout=True, timeout=100,
                              verify_certs=self.elastic.requests.verify,
                              connection_class=RequestsHttpConnection)
        es_out = Elasticsearch([enrich_backend.elastic.url], retry_on_timeout=True,
                               timeout=100, verify_certs=self.elastic.requests.verify,
                               connection_class=RequestsHttpConnection)
        in_conn = ESPandasConnector(es_conn=es_in, es_index=in_index, sort_on_field=sort_on_field)
        out_conn = ESPandasConnector(es_conn=es_out, es_index=out_index, sort_on_field=sort_on_field, read_only=False)

        exists_index = out_conn.exists()
        if no_incremental or not exists_index:
            logger.info("{} Creating out ES index".format(log_prefix))
            # Initialize out index

            if not self.elastic.is_legacy():
                filename = pkg_resources.resource_filename('grimoire_elk', 'enriched/mappings/git_aoc_es7.json')
            else:
                filename = pkg_resources.resource_filename('grimoire_elk', 'enriched/mappings/git_aoc.json')
            out_conn.create_index(filename, delete=exists_index)

        repos = []
        for source in self.json_projects.values():
            items = source.get('git')
            if items:
                repos.extend(items)

        for repo in repos:
            anonymize_repo = anonymize_url(repo)
            repo_name = anonymize_repo.split()[0]
            logger.info("{} Processing repo: {}".format(log_prefix, repo_name))
            in_conn.update_repo(repo_name)
            out_conn.update_repo(repo_name)
            areas_of_code(git_enrich=enrich_backend, in_conn=in_conn, out_conn=out_conn)

            # delete the documents in the AOC index which correspond to commits that don't exist in the raw index
            if out_conn.exists():
                self.update_items_aoc(ocean_backend, es_out, out_index, repo_name)

        # Create alias if output index exists and alias does not
        if out_conn.exists():
            if not out_conn.exists_alias(alias) \
                    and not enrich_backend.elastic.alias_in_use(alias):
                logger.info("{} creating alias: {}".format(log_prefix, alias))
                out_conn.create_alias(alias)
            else:
                logger.warning("{} alias already exists: {}.".format(log_prefix, alias))

        logger.info("{} end".format(log_prefix))

    def get_unique_hashes_aoc(self, es_aoc, index_aoc, repository):
        """Retrieve the unique commit hashes in the AOC index

        :param es_aoc: the ES object to access AOC data
        :param index_aoc: the AOC index
        :param repository: the target repository
        """

        def __unique_commit_hashes_aoc(repository, until_date=None):
            """Retrieve all unique commit hashes in ascending order on grimoire_creation_date
            for a given repository in the AOC index"""

            fltr = [
                {
                    "term": {
                        "repository": repository
                    }
                }
            ]

            if until_date:
                fltr.append({
                    "range": {
                        "metadata__updated_on": {
                            "gte": until_date
                        }
                    }
                })

            query_unique_hashes = """
            {
                "aggs": {
                    "2": {
                      "terms": {
                        "field": "hash",
                        "size": 1000,
                        "order": {
                          "1": "asc"
                        }
                      },
                      "aggs": {
                        "1": {
                          "max": {
                            "field": "grimoire_creation_date"
                          }
                        }
                      }
                    }
                },
                "size": 0,
                "query": {
                    "bool": {
                        "filter": %s
                    }
                }
            }
            """ % json.dumps(fltr)

            return query_unique_hashes

        aoc_hashes = []
        fetching = True
        last_date = None
        previous_date = None
        while fetching:
            hits = es_aoc.search(index=index_aoc, body=__unique_commit_hashes_aoc(repository, last_date))
            buckets = hits['aggregations']['2']['buckets']

            if not buckets:
                fetching = False

            for bucket in buckets:
                aoc_hashes.append(bucket['key'])
                last_date = bucket['1']['value_as_string']

            if previous_date == last_date:
                fetching = False

            previous_date = last_date

        return aoc_hashes

    def get_diff_commits_raw_aoc(self, ocean_backend, es_aoc, index_aoc, repository):
        """Return the commit hashes which are stored in the AOC index but not in the Git raw index.

        :param ocean_backend: Ocean backend
        :param es_aoc: the ES object to access AOC data
        :param index_aoc: the AOC index
        :param repository: the target repository
        """
        fltr = {
            'name': 'origin',
            'value': [repository]
        }

        raw_hashes = set([item['data']['commit']
                          for item in ocean_backend.fetch(ignore_incremental=True, _filter=fltr)])
        aoc_hashes = set(self.get_unique_hashes_aoc(es_aoc, index_aoc, repository))

        hashes_to_delete = list(aoc_hashes.difference(raw_hashes))

        return hashes_to_delete

    def update_items_aoc(self, ocean_backend, es_aoc, index_aoc, repository):
        """Update the documents stored in the AOC index by deleting those ones corresponding
        to deleted commits

        :param ocean_backend: the Ocean backend to access the raw data
        :param es_aoc: the ES object to access AOC data
        :param index_aoc: the AOC index
        :param repository: the target repository
        """
        aoc_index_url = self.elastic_url + '/' + index_aoc
        hashes_to_delete = self.get_diff_commits_raw_aoc(ocean_backend, es_aoc, index_aoc, repository)
        to_process = []
        for _hash in hashes_to_delete:
            to_process.append(_hash)

            if len(to_process) != MAX_BULK_UPDATE_SIZE:
                continue

            # delete documents from the AOC index
            self.remove_commits(to_process, aoc_index_url, 'hash', repository)

            to_process = []

        if to_process:
            # delete documents from the AOC index
            self.remove_commits(to_process, aoc_index_url, 'hash', repository)

        logger.debug("[git] study areas_of_code {} commits deleted from {} with origin {}.".format(
            len(hashes_to_delete), anonymize_url(aoc_index_url), repository))

    def enrich_onion(self, ocean_backend, enrich_backend, alias,
                     no_incremental=False,
                     in_index='git_onion-src',
                     out_index='git_onion-enriched',
                     data_source='git',
                     contribs_field='hash',
                     timeframe_field='grimoire_creation_date',
                     sort_on_field='metadata__timestamp',
                     seconds=Enrich.ONION_INTERVAL):

        super().enrich_onion(enrich_backend=enrich_backend,
                             alias=alias,
                             in_index=in_index,
                             out_index=out_index,
                             data_source=data_source,
                             contribs_field=contribs_field,
                             timeframe_field=timeframe_field,
                             sort_on_field=sort_on_field,
                             no_incremental=no_incremental,
                             seconds=seconds)

    def get_diff_commits_origin_raw(self, ocean_backend):
        """Return the commit hashes which are stored in the raw index but not in the original repo.

        :param ocean_backend: Ocean backend
        """
        repo_origin = anonymize_url(self.perceval_backend.origin)
        fltr = {
            'name': 'origin',
            'value': [repo_origin]
        }

        current_hashes = []
        try:
            git_repo = GitRepository(self.perceval_backend.uri, self.perceval_backend.gitpath)
            current_hashes = [commit for commit in git_repo.rev_list()]
        except EmptyRepositoryError:
            logger.warning("No commits retrieved from {}, repo is empty".format(repo_origin))
        except RepositoryError:
            logger.warning("No commits retrieved from {}, repo doesn't exist locally".format(repo_origin))
        except Exception as e:
            logger.error("[git] No commits retrieved from {}, git rev-list command failed: {}".format(repo_origin, e))

        if not current_hashes:
            return current_hashes

        current_hashes = set(current_hashes)
        raw_hashes = set([item['data']['commit']
                          for item in ocean_backend.fetch(ignore_incremental=True, _filter=fltr)])

        hashes_to_delete = list(raw_hashes.difference(current_hashes))

        return hashes_to_delete

    def update_items(self, ocean_backend, enrich_backend):
        """Retrieve the commits not present in the original repository and delete
        the corresponding documents from the raw and enriched indexes"""

        repo_origin = anonymize_url(self.perceval_backend.origin)
        logger.debug("[git] update-items Checking commits for {}.".format(repo_origin))
        hashes_to_delete = self.get_diff_commits_origin_raw(ocean_backend)

        to_process = []
        for _hash in hashes_to_delete:
            to_process.append(_hash)

            if len(to_process) != MAX_BULK_UPDATE_SIZE:
                continue

            # delete documents from the raw index
            self.remove_commits(to_process, ocean_backend.elastic.index_url, 'data.commit', repo_origin)
            # delete documents from the enriched index
            self.remove_commits(to_process, enrich_backend.elastic.index_url, 'hash', repo_origin)

            to_process = []

        if to_process:
            # delete documents from the raw index
            self.remove_commits(to_process, ocean_backend.elastic.index_url, 'data.commit', repo_origin)
            # delete documents from the enriched index
            self.remove_commits(to_process, enrich_backend.elastic.index_url, 'hash', repo_origin)

        logger.debug("[git] update-items {} commits deleted from {} with origin {}.".format(
                     len(hashes_to_delete), anonymize_url(ocean_backend.elastic.index_url),
                     repo_origin))
        logger.debug("[git] update-items {} commits deleted from {} with origin {}.".format(
                     len(hashes_to_delete), anonymize_url(enrich_backend.elastic.index_url),
                     repo_origin))

    def remove_commits(self, items, index, attr, origin, origin_attr='origin'):
        """Delete documents that correspond to commits deleted in the Git repository

        :param items: target items to be deleted
        :param index: target index
        :param attr: name of the term attribute to search items
        :param origin: name of the origin from where the items must be deleted
        :param origin_attr: attribute where the origin info is stored.
        """
        es_query = '''
            {
              "query": {
                "bool": {
                    "must": {
                        "term": {
                            "%s": "%s"
                        }
                    },
                    "filter": {
                        "terms": {
                            "%s": [%s]
                        }
                    }
                }
              }
            }
            ''' % (origin_attr, origin, attr, ",".join(['"%s"' % i for i in items]))

        r = self.requests.post(index + "/_delete_by_query?refresh", data=es_query, headers=HEADER_JSON, verify=False)
        try:
            r.raise_for_status()
        except requests.exceptions.HTTPError as ex:
            logger.error("[git] Error updating deleted commits for {}.".format(anonymize_url(index)))
            logger.error(r.text)
            return

    def enrich_git_branches(self, ocean_backend, enrich_backend, run_month_days=[7, 14, 21, 28]):
        """Update the information about branches within the documents representing
        commits in the enriched index.

        The example below shows how to activate the study by modifying the setup.cfg. The study
        `enrich_git_branches` will be run on days depending on the parameter `run_month_days`,
        by default the days are 7, 14, 21, and 28 of each month.

        ```
        [git]
        raw_index = git_raw
        enriched_index = git_enriched
        ...
        studies = [enrich_git_branches]

        [enrich_git_branches]
        run_month_days = [5, 22]
        ```

        :param ocean_backend: the ocean backend
        :param enrich_backend: the enrich backend
        :param run_month_days: days of the month to run this study
        """
        logger.info("[git] study git-branches start")
        day = datetime_utcnow().day
        run_month_days = list(map(int, run_month_days))
        if day not in run_month_days:
            logger.info("[git] study git-branches will execute only the days {} of each month".format(run_month_days))
            logger.info("[git] study git-branches end")
            return

        for ds in self.prjs_map:
            if ds != "git":
                continue

            urls = self.prjs_map[ds]

            for url in urls:
                if '--filter-no-collection=true' in url:
                    # Skip study when --filter-no-collection is present
                    logger.info("[git] study git-branches skipping repo {}".format(anonymize_url(url)))
                    continue
                cmd = GitCommand(*[url])
                try:
                    git_repo = GitRepository(cmd.parsed_args.uri, cmd.parsed_args.gitpath)
                except RepositoryError:
                    logger.error("[git] study git-branches skipping not cloned repo {}".format(anonymize_url(url)))
                    continue

                logger.debug("[git] study git-branches delete branch info for repo {} in index {}".format(
                             git_repo.uri, anonymize_url(enrich_backend.elastic.index_url)))
                self.delete_commit_branches(git_repo, enrich_backend)

                logger.debug("[git] study git-branches add branch info for repo {} in index {}".format(
                             git_repo.uri, anonymize_url(enrich_backend.elastic.index_url)))
                try:
                    self.add_commit_branches(git_repo, enrich_backend)
                except Exception as e:
                    logger.error("[git] study git-branches failed on repo {}, due to {}".format(git_repo.uri, e))
                    continue

                logger.debug("[git] study git-branches repo {} in index {} processed".format(
                             git_repo.uri, anonymize_url(enrich_backend.elastic.index_url)))

        logger.info("[git] study git-branches end")

    def delete_commit_branches(self, git_repo, enrich_backend):
        """Delete the information about branches from the documents representing
        commits in the enriched index.

        :param git_repo: GitRepository object
        :param enrich_backend: the enrich backend
        """
        fltr = """
            "filter": [
                {
                    "term": {
                        "origin": "%s"
                    }
                }
            ]
        """ % anonymize_url(git_repo.uri)

        # reset references in enrich index
        es_query = """
            {
              "script": {
                "source": "ctx._source.branches = new HashSet();",
                "lang": "painless"
              },
              "query": {
                "bool": {
                    %s
                }
              }
            }
            """ % fltr

        index = enrich_backend.elastic.index_url
        r = self.requests.post(index + "/_update_by_query?refresh", data=es_query, headers=HEADER_JSON, verify=False)
        try:
            r.raise_for_status()
        except requests.exceptions.HTTPError:
            logger.error("[git] Error while deleting branches on {}".format(anonymize_url(index)))
            logger.error(r.text)
            return

        logger.debug("[git] Delete branches {}, index {}".format(r.text, anonymize_url(index)))

    def add_commit_branches(self, git_repo, enrich_backend):
        """Add the information about branches to the documents representing commits in
        the enriched index. Branches are obtained using the command `git ls-remote`,
        then for each branch, the list of commits is retrieved via the command `git rev-list branch-name` and
        used to update the corresponding items in the enriched index.

        :param git_repo: GitRepository object
        :param enrich_backend: the enrich backend
        """
        to_process = []
        for hash, refname in git_repo._discover_refs(remote=True):

            if not refname.startswith('refs/heads/'):
                continue

            commit_count = 0
            branch_name = refname.replace('refs/heads/', '')

            try:
                commits = git_repo.rev_list([branch_name])

                for commit in commits:
                    to_process.append(commit)
                    commit_count += 1

                    if commit_count == MAX_BULK_UPDATE_SIZE:
                        self.__process_commits_in_branch(enrich_backend, git_repo.uri, branch_name, to_process)

                        # reset the counter
                        to_process = []
                        commit_count = 0

                if commit_count:
                    self.__process_commits_in_branch(enrich_backend, git_repo.uri, branch_name, to_process)

            except Exception as e:
                logger.error("[git] Skip adding branch info for repo {} due to {}".format(git_repo.uri, e))
                return

    def __process_commits_in_branch(self, enrich_backend, repo_origin, branch_name, commits):
        commits_str = ",".join(['"%s"' % c for c in commits])

        # process branch names which include quotes or single quote
        digested_branch_name = branch_name
        if "'" in branch_name:
            digested_branch_name = branch_name.replace("'", "---")
            logger.warning("[git] Change branch name from {} to {}".format(branch_name, digested_branch_name))
        if '"' in branch_name:
            digested_branch_name = branch_name.replace('"', "---")
            logger.warning("[git] Change branch name from {} to {}".format(branch_name, digested_branch_name))

        # update enrich index
        fltr = self.__prepare_filter("hash", commits_str, anonymize_url(repo_origin))

        es_query = """
            {
              "script": {
                "source": "if(!ctx._source.branches.contains(params.branch)){ctx._source.branches.add(params.branch);}",
                "lang": "painless",
                "params": {
                    "branch": "'%s'"
                }
              },
              "query": {
                "bool": {
                    %s
                }
              }
            }
            """ % (digested_branch_name, fltr)

        index = enrich_backend.elastic.index_url
        r = self.requests.post(index + "/_update_by_query?refresh", data=es_query, headers=HEADER_JSON, verify=False)
        try:
            r.raise_for_status()
        except requests.exceptions.HTTPError:
            logger.error("[git] Error adding branch info for {}".format(anonymize_url(index)))
            logger.error(r.text)
            return

        logger.debug("[git] Add branches {}, index {}".format(r.text, anonymize_url(index)))

    def __prepare_filter(self, terms_attr, terms_value, repo_origin):
        fltr = """
            "filter": [
                {
                    "terms": {
                        "%s": [%s]
                    }
                },
                {
                    "term": {
                        "origin": "%s"
                    }
                }
            ]
        """ % (terms_attr, terms_value, repo_origin)

        return fltr



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2023 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#

import datetime
import inspect
import json
import logging
import re

import requests
import urllib3

from grimoirelab_toolkit.datetime import (datetime_utcnow,
                                          str_to_datetime)


BACKOFF_FACTOR = 0.2
MAX_RETRIES = 21
MAX_RETRIES_ON_REDIRECT = 5
MAX_RETRIES_ON_READ = 8
MAX_RETRIES_ON_CONNECT = 21
STATUS_FORCE_LIST = [408, 409, 429, 502, 503, 504]
METADATA_FILTER_RAW = 'metadata__filter_raw'
REPO_LABELS = 'repository_labels'

logger = logging.getLogger(__name__)


def get_repository_filter(perceval_backend, perceval_backend_name, term=False):
    """ Get the filter needed for get the items in a repository """
    from .github import GITHUB

    filter_ = {}

    if not perceval_backend:
        return filter_

    field = 'origin'
    value = anonymize_url(perceval_backend.origin)

    if perceval_backend_name in ["meetup", "nntp", "stackexchange", "jira", "hyperkitty"]:
        # Until tag is supported in all raw and enriched indexes
        # we should use origin. But stackexchange and meetup won't work with origin
        # because the tag must be included in the filter.
        # For nntp we have a common group server as origin, so we need to use also the tag.
        # And in jira we can filter by product, and the origin is the same jira server.
        field = 'tag'
        value = perceval_backend.tag

    if perceval_backend:
        if not term:
            filter_ = {"name": field,
                       "value": value}
        else:
            filter_ = '''
                {"term":
                    { "%s" : "%s"  }
                }
            ''' % (field, value)
            # Filters are always a dict
            filter_ = json.loads(filter_)

    if value in ['', GITHUB + '/', 'https://meetup.com/']:
        # Support for getting all items from a multiorigin index
        # In GitHub we receive GITHUB + '/', the site url without org and repo
        # In Meetup we receive https://meetup.com/ as the tag
        filter_ = {}

    return filter_


def get_confluence_spaces_filter(repo_spaces, perceval_backend_name):
    """ Get the spaces needed for get the items in a confluence instance """

    filter_ = {}

    if not repo_spaces or perceval_backend_name != "confluence":
        return filter_

    field = 'data._expandable.space'
    value_path = "/rest/api/space/"
    values = repo_spaces

    filter_ = {
        "should": []
    }

    for value in values:
        new = {
            "term": {field: value_path + value}
        }
        filter_["should"].append(new)

    return filter_


def anonymize_url(url):
    """Remove credentials from the url

    :param url: target url
    """
    anonymized = re.sub('://.*@', '://', url)

    return anonymized


def get_time_diff_days(start, end):
    ''' Number of days between two dates in UTC format  '''

    if start is None or end is None:
        return None

    if type(start) is not datetime.datetime:
        start = str_to_datetime(start).replace(tzinfo=None)
    if type(end) is not datetime.datetime:
        end = str_to_datetime(end).replace(tzinfo=None)

    seconds_day = float(60 * 60 * 24)
    diff_days = (end - start).total_seconds() / seconds_day
    diff_days = float('%.2f' % diff_days)

    return diff_days


def grimoire_con(insecure=True, conn_retries=MAX_RETRIES_ON_CONNECT, total=MAX_RETRIES):
    conn = requests.Session()
    # {backoff factor} * (2 ^ ({number of total retries} - 1))
    # conn_retries = 21  # 209715.2 = 2.4d
    # total covers issues like 'ProtocolError('Connection aborted.')
    # Retry when there are errors in HTTP connections
    retries = urllib3.util.Retry(total=total, connect=conn_retries, read=MAX_RETRIES_ON_READ,
                                 redirect=MAX_RETRIES_ON_REDIRECT, backoff_factor=BACKOFF_FACTOR,
                                 allowed_methods=False, status_forcelist=STATUS_FORCE_LIST)
    adapter = requests.adapters.HTTPAdapter(max_retries=retries)
    conn.mount('http://', adapter)
    conn.mount('https://', adapter)

    if insecure:
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
        conn.verify = False

    return conn


def get_last_enrich(backend_cmd, enrich_backend, filter_raw=None):
    last_enrich = None

    if backend_cmd:
        backend = backend_cmd.backend
        # Only supported in data retrieved from a perceval backend
        # Always filter by repository to support multi repository indexes
        backend_name = enrich_backend.get_connector_name()

        filter_ = get_repository_filter(backend, backend_name)
        filters_ = [filter_]

        # include the filter_raw text (taken from the projects.json) to the
        # list of filters used to retrieve the last enrich date.
        if filter_raw:
            filter_raw_ = {
                "name": METADATA_FILTER_RAW,
                "value": filter_raw
            }

            filters_.append(filter_raw_)

        # Check if backend supports from_date
        signature = inspect.signature(backend.fetch)

        from_date = None
        if 'from_date' in signature.parameters:
            try:
                # Support perceval pre and post BackendCommand refactoring
                from_date = backend_cmd.from_date
            except AttributeError:
                from_date = backend_cmd.parsed_args.from_date

        offset = None
        if 'offset' in signature.parameters:
            try:
                offset = backend_cmd.offset
            except AttributeError:
                offset = backend_cmd.parsed_args.offset

        if from_date:
            if from_date.replace(tzinfo=None) != str_to_datetime("1970-01-01").replace(tzinfo=None):
                last_enrich = from_date
            # if the index is empty, set the last enrich to None
            elif not enrich_backend.from_date:
                last_enrich = None
            else:
                # if the index is not empty, the last enrich is the minimum between
                # the last filtered item (the last item for a given origin) and the
                # last item in the enriched index. If `last_enrich_filtered` is empty,
                # it means that no items for that origin are in the index, thus the
                # `last_enrich` is set to None
                last_enrich_filtered = enrich_backend.get_last_update_from_es(filters_)
                last_enrich = get_min_last_enrich(enrich_backend.from_date, last_enrich_filtered)

        elif offset is not None:
            if offset != 0:
                last_enrich = offset
            else:
                last_enrich = enrich_backend.get_last_offset_from_es(filters_)

        else:
            if not enrich_backend.from_date:
                last_enrich = None
            else:
                last_enrich_filtered = enrich_backend.get_last_update_from_es(filters_)
                last_enrich = get_min_last_enrich(enrich_backend.from_date, last_enrich_filtered)
    else:
        last_enrich = enrich_backend.get_last_update_from_es()

    return last_enrich


def get_min_last_enrich(last_enrich, last_enrich_filtered):
    if last_enrich_filtered:
        min_enrich = min(last_enrich, last_enrich_filtered.replace(tzinfo=None))
    else:
        min_enrich = None

    return min_enrich


def get_diff_current_date(days=0, hours=0, minutes=0):
    before_date = datetime_utcnow() - datetime.timedelta(days=days, hours=hours, minutes=minutes)

    return before_date


def fix_field_date(date_value):
    """Fix possible errors in the field date"""

    field_date = str_to_datetime(date_value)

    try:
        _ = int(field_date.strftime("%z")[0:3])
    except ValueError:
        field_date = field_date.replace(tzinfo=None)

    return field_date.isoformat()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2023 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jose Javier Merchante Picazo <jjmerchante@cauldron.io>
#

from grimoire_elk.identities.identities import Identities


class MeetupIdentities(Identities):
    @classmethod
    def anonymize_item(cls, item):
        """Remove or hash the fields that contain personal information"""

        item = item['data']

        if 'event_hosts' in item:
            for i, host in enumerate(item['event_hosts']):
                item['event_hosts'][i] = {
                    'id': host['id'],
                    'name': cls._hash(host['name']),
                }

        if 'rsvps' in item:
            for rsvp in item['rsvps']:
                rsvp['member'] = {
                    'id': rsvp['member']['id'],
                    'name': cls._hash(rsvp['member']['name']),
                    'event_context': {'host': rsvp['member']['event_context']['host']}
                }

        if 'comments' in item:
            for comment in item['comments']:
                comment['member'] = {
                    'id': comment['member']['id'],
                    'name': cls._hash(comment['member']['name'])
                }



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2023 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#

from .enrich import Enrich, metadata
from grimoirelab_toolkit.datetime import str_to_datetime


class PuppetForgeEnrich(Enrich):

    def get_elastic_mappings(self):

        mapping = """
        {
            "properties": {
                "summary_analyzed": {
                  "type": "text",
                  "index":"true"
                }
           }
        } """

        return {"items": mapping}

    def get_identities(self, item):
        """ Return the identities from an item """

        user = self.get_sh_identity(item, self.get_field_author())
        yield user

        # Get the identities from the releases
        for release in item['data']['releases']:
            user = self.get_sh_identity(release['module'], self.get_field_author())
            yield user

    def get_field_author(self):
        return 'owner'

    def get_sh_identity(self, item, identity_field=None):

        entry = item

        if isinstance(item, dict) and 'data' in item:
            entry = item['data']

        identity = {f: None for f in ['email', 'name', 'username']}

        if identity_field in entry:
            identity['username'] = entry[identity_field]['username']

        return identity

    def get_field_event_unique_id(self):
        """ Field in the rich event with the unique id """
        return "uuid"

    @metadata
    def get_rich_item(self, item):
        eitem = {}

        self.copy_raw_fields(self.RAW_FIELDS_COPY, item, eitem)
        # The real data
        entry = item['data']

        # data fields to copy
        copy_fields = ["downloads", "uri", "name", "slug", "issues_url",
                       "homepage_url"]
        for f in copy_fields:
            if f in entry:
                eitem[f] = entry[f]
            else:
                eitem[f] = None
        # Fields which names are translated
        map_fields = {"feedback_score": "module_feedback_score",
                      "uri": "module_uri",
                      "name": "module_name",
                      "slug": "module_slug",
                      "downloads": "module_downloads"}
        for f in map_fields:
            if f in entry:
                eitem[map_fields[f]] = entry[f]

        eitem["is_release"] = 0
        eitem["type"] = "module"
        eitem["url"] = 'https://forge.puppet.com/' + entry['owner']['username'] + '/' + entry['name']
        eitem["author_url"] = 'https://forge.puppet.com/' + entry['owner']['username']
        eitem["gravatar_id"] = entry['owner']['gravatar_id']
        eitem["owner_release_count"] = entry['owner_data']['release_count']
        eitem["owner_module_count"] = entry['owner_data']['module_count']
        eitem["module_downloads"] = entry['downloads']
        eitem["release_count"] = len(entry['releases'])
        if 'current_release' in entry:
            eitem["module_version"] = entry['current_release']['version']
            eitem["version"] = entry['current_release']['version']
            eitem["validation_score"] = entry['current_release']['validation_score']
            eitem["tags"] = entry['current_release']['tags']
            eitem["license"] = entry['current_release']['metadata']['license']
            eitem["source_url"] = entry['current_release']['metadata']['source']
            eitem["summary"] = entry['current_release']['metadata']['summary']

        if self.sortinghat:
            eitem.update(self.get_item_sh(item))

        if self.prjs_map:
            eitem.update(self.get_item_project(eitem))

        eitem.update(self.get_grimoire_fields(entry["created_at"], "module"))

        self.add_repository_labels(eitem)
        self.add_metadata_filter_raw(eitem)
        return eitem

    def get_rich_events(self, item):
        """
        Get the enriched events related to a module
        """
        module = item['data']
        if not item['data']['releases']:
            return []

        for release in item['data']['releases']:
            event = self.get_rich_item(item)
            # Update specific fields for this release
            event["uuid"] += "_" + release['slug']
            event["author_url"] = 'https://forge.puppet.com/' + release['module']['owner']['username']
            event["gravatar_id"] = release['module']['owner']['gravatar_id']
            event["downloads"] = release['downloads']
            event["slug"] = release['slug']
            event["version"] = release['version']
            event["uri"] = release['uri']
            event["validation_score"] = release['validation_score']
            event["homepage_url"] = None
            if 'project_page' in release['metadata']:
                event["homepage_url"] = release['metadata']['project_page']
            event["issues_url"] = None
            if "issues_url" in release['metadata']:
                event["issues_url"] = release['metadata']['issues_url']
            event["tags"] = release['tags']
            event["license"] = release['metadata']['license']
            event["source_url"] = release['metadata']['source']
            event["summary"] = release['metadata']['summary']

            event["metadata__updated_on"] = str_to_datetime(release['updated_at']).isoformat()

            if self.sortinghat:
                release["metadata__updated_on"] = event["metadata__updated_on"]  # Needed in get_item_sh logic
                event.update(self.get_item_sh(release))

            if self.prjs_map:
                event.update(self.get_item_project(event))

            event.update(self.get_grimoire_fields(release["created_at"], "release"))

            yield event

    def enrich_items(self, ocean_backend):
        total = super(PuppetForgeEnrich, self).enrich_items(ocean_backend)
        # Always generate also the events
        super(PuppetForgeEnrich, self).enrich_items(ocean_backend, events=True)

        return total  # return just the enriched items, not the events



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2023 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Alvaro del Castillo <acs@bitergia.com>
#     Valerio Cosentino <valcos@bitergia.com>
#     Miguel Ángel Fernández <mafesan@bitergia.com>
#
import logging
import time
import unittest

import requests

from base import TestBaseBackend
from grimoire_elk.enriched.enrich import logger
from grimoire_elk.enriched.github import logger as logger_github
from grimoire_elk.enriched.utils import REPO_LABELS, anonymize_url
from grimoire_elk.raw.github import GitHubOcean

HEADER_JSON = {"Content-Type": "application/json"}


class TestGitHub(TestBaseBackend):
    """Test GitHub backend"""

    connector = "github"
    ocean_index = "test_" + connector
    enrich_index = "test_" + connector + "_enrich"
    ocean_index_anonymized = "test_" + connector + "_anonymized"
    enrich_index_anonymized = "test_" + connector + "_enrich_anonymized"

    def test_has_identites(self):
        """Test value of has_identities method"""

        enrich_backend = self.connectors[self.connector][2]()
        self.assertTrue(enrich_backend.has_identities())

    def test_items_to_raw(self):
        """Test whether JSON items are properly inserted into ES"""

        result = self._test_items_to_raw()

        self.assertGreater(result['items'], 0)
        self.assertGreater(result['raw'], 0)
        self.assertEqual(result['items'], result['raw'])

    def test_raw_to_enrich(self):
        """Test whether the raw index is properly enriched"""

        result = self._test_raw_to_enrich()

        self.assertGreater(result['raw'], 0)
        self.assertGreater(result['enrich'], 0)
        self.assertEqual(result['raw'], result['enrich'])

        enrich_backend = self.connectors[self.connector][2]()

        item = self.items[0]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(eitem['labels'], [])
        self.assertEqual(item['category'], 'issue')
        self.assertEqual(eitem['author_uuid'], '5f9d42ce000e46e9eee60a3c64a353b560051a2e')
        self.assertEqual(eitem['author_domain'], 'zhquan_example.com')
        self.assertEqual(eitem['user_data_uuid'], '5f9d42ce000e46e9eee60a3c64a353b560051a2e')
        self.assertEqual(eitem['user_data_domain'], 'zhquan_example.com')
        self.assertEqual(eitem['assignee_data_uuid'], '5f9d42ce000e46e9eee60a3c64a353b560051a2e')
        self.assertEqual(eitem['assignee_data_domain'], 'zhquan_example.com')

        self.assertEqual(eitem['url'], 'https://github.com/zhquan_example/repo/pull/1')
        self.assertEqual(eitem['issue_url'], 'https://github.com/zhquan_example/repo/pull/1')

        item = self.items[1]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(eitem['labels'], ['bug', 'feature'])
        self.assertEqual(item['category'], 'pull_request')
        self.assertEqual(eitem['time_to_merge_request_response'], 335.81)
        self.assertEqual(eitem['author_uuid'], '5f9d42ce000e46e9eee60a3c64a353b560051a2e')
        self.assertEqual(eitem['author_domain'], 'zhquan_example.com')
        self.assertEqual(eitem['user_data_uuid'], '5f9d42ce000e46e9eee60a3c64a353b560051a2e')
        self.assertEqual(eitem['user_data_domain'], 'zhquan_example.com')
        self.assertEqual(eitem['merged_by_data_uuid'], '5f9d42ce000e46e9eee60a3c64a353b560051a2e')
        self.assertEqual(eitem['merged_by_data_domain'], 'zhquan_example.com')
        self.assertEqual(eitem['additions'], 528)
        self.assertEqual(eitem['deletions'], 0)
        self.assertEqual(eitem['changed_files'], 4)

        self.assertEqual(eitem['url'], 'https://github.com/zhquan_example/repo/pull/1')
        self.assertEqual(eitem['issue_url'], 'https://github.com/zhquan_example/repo/pull/1')

        item = self.items[2]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(item['category'], 'repository')
        self.assertEqual(eitem['forks_count'], 16687)
        self.assertEqual(eitem['subscribers_count'], 2904)
        self.assertEqual(eitem['stargazers_count'], 48188)
        self.assertEqual(eitem['url'], "https://github.com/kubernetes/kubernetes")

        item = self.items[3]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(item['category'], 'repository')
        self.assertEqual(eitem['forks_count'], 16687)
        self.assertEqual(eitem['subscribers_count'], 4301)
        self.assertEqual(eitem['stargazers_count'], 47118)
        self.assertEqual(eitem['url'], "https://github.com/kubernetes/kubernetes")

        item = self.items[4]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(item['category'], 'repository')
        self.assertEqual(eitem['forks_count'], 1)
        self.assertEqual(eitem['subscribers_count'], 1)
        self.assertEqual(eitem['stargazers_count'], 1)
        self.assertEqual(eitem['url'], "https://github.com/kubernetes/kubernetes")

        item = self.items[5]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(item['category'], 'issue')
        self.assertEqual(eitem['user_name'], 'acs')
        self.assertIsNone(eitem['user_domain'])
        self.assertIsNone(eitem['user_org'])
        self.assertEqual(eitem['author_name'], 'acs')
        self.assertEqual(eitem['author_uuid'], 'e8cc482634f2095c935b6a586ddb9ed8215d5cb8')
        self.assertIsNone(eitem['assignee_name'])
        self.assertIsNone(eitem['assignee_domain'])
        self.assertIsNone(eitem['assignee_org'])
        self.assertEqual(eitem['user_data_name'], 'acs')
        self.assertEqual(eitem['user_data_uuid'], 'e8cc482634f2095c935b6a586ddb9ed8215d5cb8')
        self.assertIsNone(eitem['user_data_domain'])

        self.assertEqual(eitem['url'], 'https://github.com/chaoss/grimoirelab-perceval/pull/7')
        self.assertEqual(eitem['issue_url'], 'https://github.com/chaoss/grimoirelab-perceval/pull/7')

        item = self.items[6]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(item['category'], 'pull_request')
        self.assertEqual(eitem['user_name'], 'acs')
        self.assertIsNone(eitem['user_domain'])
        self.assertIsNone(eitem['user_org'])
        self.assertEqual(eitem['author_name'], 'acs')
        self.assertEqual(eitem['author_uuid'], 'e8cc482634f2095c935b6a586ddb9ed8215d5cb8')
        self.assertIsNone(eitem['merge_author_name'])
        self.assertIsNone(eitem['merge_author_domain'])
        self.assertIsNone(eitem['merge_author_org'])
        self.assertEqual(eitem['user_data_name'], 'acs')
        self.assertEqual(eitem['user_data_uuid'], 'e8cc482634f2095c935b6a586ddb9ed8215d5cb8')
        self.assertIsNone(eitem['user_data_domain'])
        self.assertEqual(eitem['additions'], 5)
        self.assertEqual(eitem['deletions'], 1)
        self.assertEqual(eitem['changed_files'], 1)

        self.assertEqual(eitem['url'], 'https://github.com/chaoss/grimoirelab-perceval/pull/4')
        self.assertEqual(eitem['issue_url'], 'https://github.com/chaoss/grimoirelab-perceval/pull/4')

    def test_enrich_repo_labels(self):
        """Test whether the field REPO_LABELS is present in the enriched items"""

        self._test_raw_to_enrich()
        enrich_backend = self.connectors[self.connector][2]()

        for item in self.items:
            eitem = enrich_backend.get_rich_item(item)
            self.assertIn(REPO_LABELS, eitem)

    def test_raw_to_enrich_sorting_hat(self):
        """Test enrich with SortingHat"""

        result = self._test_raw_to_enrich(sortinghat=True)
        self.assertGreater(result['raw'], 0)
        self.assertGreater(result['enrich'], 0)
        self.assertEqual(result['raw'], result['enrich'])

        enrich_backend = self.connectors[self.connector][2]()

        url = self.es_con + "/" + self.enrich_index + "/_search"
        response = enrich_backend.requests.get(url, verify=False).json()
        for hit in response['hits']['hits']:
            source = hit['_source']
            if 'author_uuid' in source:
                self.assertIn('author_domain', source)
                self.assertIn('author_gender', source)
                self.assertIn('author_gender_acc', source)
                self.assertIn('author_org_name', source)
                self.assertIn('author_bot', source)
                self.assertIn('author_multi_org_names', source)

    def test_raw_to_enrich_projects(self):
        """Test enrich with Projects"""

        result = self._test_raw_to_enrich(projects=True)
        # ... ?

    def test_refresh_identities(self):
        """Test refresh identities"""

        result = self._test_refresh_identities()
        # ... ?

    def test_perceval_params(self):
        """Test the extraction of perceval params from an URL"""

        url = "https://github.com/chaoss/grimoirelab-perceval"
        expected_params = [
            'chaoss', 'grimoirelab-perceval'
        ]
        self.assertListEqual(GitHubOcean.get_perceval_params_from_url(url), expected_params)

    def test_demography_study(self):
        """ Test that the demography study works correctly """

        alias = "demographics"
        study, ocean_backend, enrich_backend = self._test_study('enrich_demography')

        with self.assertLogs(logger, level='INFO') as cm:

            if study.__name__ == "enrich_demography":
                study(ocean_backend, enrich_backend, alias)

            self.assertEqual(cm.output[0], 'INFO:grimoire_elk.enriched.enrich:[github] Demography '
                                           'starting study %s/test_github_enrich'
                             % anonymize_url(self.es_con))
            self.assertEqual(cm.output[-1], 'INFO:grimoire_elk.enriched.enrich:[github] Demography '
                                            'end %s/test_github_enrich'
                             % anonymize_url(self.es_con))

        time.sleep(5)  # HACK: Wait until github enrich index has been written
        items = [item for item in enrich_backend.fetch()]
        self.assertEqual(len(items), 7)
        for item in items:
            self.assertNotIn('username:password', item['origin'])
            self.assertNotIn('username:password', item['tag'])
            if 'author_uuid' in item:
                self.assertTrue('demography_min_date' in item.keys())
                self.assertTrue('demography_max_date' in item.keys())

        r = enrich_backend.elastic.requests.get(enrich_backend.elastic.index_url + "/_alias",
                                                headers=HEADER_JSON, verify=False)
        self.assertIn(alias, r.json()[enrich_backend.elastic.index]['aliases'])

    def test_geolocation_study(self):
        """ Test that the geolocation study works correctly """

        study, ocean_backend, enrich_backend = self._test_study('enrich_geolocation')

        with self.assertLogs(logger, level='INFO') as cm:

            if study.__name__ == "enrich_geolocation":
                study(ocean_backend, enrich_backend,
                      location_field="user_location", geolocation_field="user_geolocation")

            self.assertEqual(cm.output[0], 'INFO:grimoire_elk.enriched.enrich:[github] Geolocation '
                                           'starting study %s/test_github_enrich'
                             % anonymize_url(self.es_con))
            self.assertEqual(cm.output[-1], 'INFO:grimoire_elk.enriched.enrich:[github] Geolocation '
                                            'end %s/test_github_enrich'
                             % anonymize_url(self.es_con))

        time.sleep(5)  # HACK: Wait until github enrich index has been written
        items = [item for item in enrich_backend.fetch() if 'user_location' in item]
        self.assertEqual(len(items), 4)
        for item in items:
            self.assertIn('user_geolocation', item)

    def test_enrich_backlog_analysis(self):
        """ Test that the backlog analysis works correctly """

        study, ocean_backend, enrich_backend = self._test_study('enrich_backlog_analysis')

        with self.assertLogs(logger_github, level='INFO') as cm:

            if study.__name__ == "enrich_backlog_analysis":
                study(ocean_backend, enrich_backend)

            self.assertEqual(cm.output[0], 'INFO:grimoire_elk.enriched.github:[github] '
                                           'Start enrich_backlog_analysis study')
            self.assertEqual(cm.output[-1], 'INFO:grimoire_elk.enriched.github:[github] '
                                            'End enrich_backlog_analysis study')

        time.sleep(5)  # HACK: Wait until github enrich index has been written
        url = self.es_con + "/github_enrich_backlog/_search"
        response = enrich_backend.requests.get(url, verify=False).json()
        for hit in response['hits']['hits']:
            source = hit['_source']
            self.assertIn('uuid', source)
            self.assertIn('opened', source)
            self.assertIn('average_opened_time', source)
            self.assertIn('origin', source)
            self.assertIn('labels', source)
            self.assertIn('project', source)
            self.assertIn('interval_days', source)
            self.assertIn('study_creation_date', source)
            self.assertIn('metadata__enriched_on', source)
            self.assertIn('grimoire_creation_date', source)
            self.assertIn('is_github_stats', source)
            self.assertIn('organization', source)

    def test_items_to_raw_anonymized(self):
        """Test whether JSON items are properly inserted into ES anonymized"""

        result = self._test_items_to_raw_anonymized()

        self.assertGreater(result['items'], 0)
        self.assertGreater(result['raw'], 0)
        self.assertEqual(result['items'], result['raw'])

        item = self.items[0]['data']
        self.assertEqual(item['assignee']['login'], '176eee2bd2c010d02dd419c453aca854195de172')
        self.assertEqual(item['assignee_data']['login'], '176eee2bd2c010d02dd419c453aca854195de172')
        self.assertEqual(item['assignee_data']['name'], '176eee2bd2c010d02dd419c453aca854195de172')
        self.assertEqual(item['user']['login'], '176eee2bd2c010d02dd419c453aca854195de172')
        self.assertEqual(item['user_data']['login'], '176eee2bd2c010d02dd419c453aca854195de172')
        self.assertEqual(item['user_data']['name'], '176eee2bd2c010d02dd419c453aca854195de172')

        item = self.items[5]['data']
        self.assertEqual(item['comments_data'][0]['user']['login'], '2283e7d3eb1195c11d3ffafe7831f94a4b5952b2')
        self.assertEqual(item['comments_data'][0]['user_data']['login'], '2283e7d3eb1195c11d3ffafe7831f94a4b5952b2')
        self.assertEqual(item['comments_data'][0]['user_data']['name'], '2283e7d3eb1195c11d3ffafe7831f94a4b5952b2')
        self.assertEqual(item['comments_data'][1]['user']['login'], '257c699509389edc61f79193ca52f2b2368a126f')
        self.assertEqual(item['comments_data'][1]['user_data']['login'], '257c699509389edc61f79193ca52f2b2368a126f')
        self.assertEqual(item['comments_data'][1]['user_data']['name'], '257c699509389edc61f79193ca52f2b2368a126f')

        item = self.items[7]['data']
        self.assertEqual(item['merged_by']['login'], '29eb8410db7377c926a7b4b8006536049df4ead8')
        self.assertEqual(item['merged_by_data']['login'], '29eb8410db7377c926a7b4b8006536049df4ead8')
        self.assertEqual(item['merged_by_data']['name'], '29eb8410db7377c926a7b4b8006536049df4ead8')
        self.assertEqual(item['user']['login'], 'b45c6541a53918180164f1ff03c9995b69456b9b')
        self.assertEqual(item['user_data']['login'], 'b45c6541a53918180164f1ff03c9995b69456b9b')
        self.assertEqual(item['user_data']['name'], 'b45c6541a53918180164f1ff03c9995b69456b9b')

    def test_raw_to_enrich_anonymized(self):
        """Test whether the raw index is properly enriched"""

        result = self._test_raw_to_enrich_anonymized()

        self.assertGreater(result['raw'], 0)
        self.assertGreater(result['enrich'], 0)
        self.assertEqual(result['raw'], result['enrich'])

        enrich_backend = self.connectors[self.connector][2]()

        item = self.items[0]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(eitem['author_uuid'], 'e4992881f28cf3318a566f0bd45dcf435216a82f')
        self.assertEqual(eitem['user_data_uuid'], 'e4992881f28cf3318a566f0bd45dcf435216a82f')
        self.assertEqual(eitem['user_data_name'], '176eee2bd2c010d02dd419c453aca854195de172')
        self.assertEqual(eitem['assignee_data_uuid'], 'e4992881f28cf3318a566f0bd45dcf435216a82f')
        self.assertEqual(eitem['assignee_data_name'], '176eee2bd2c010d02dd419c453aca854195de172')

        item = self.items[1]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(eitem['author_uuid'], 'e4992881f28cf3318a566f0bd45dcf435216a82f')
        self.assertEqual(eitem['user_data_uuid'], 'e4992881f28cf3318a566f0bd45dcf435216a82f')
        self.assertEqual(eitem['user_data_name'], '176eee2bd2c010d02dd419c453aca854195de172')
        self.assertEqual(eitem['merged_by_data_uuid'], 'e4992881f28cf3318a566f0bd45dcf435216a82f')
        self.assertEqual(eitem['merged_by_data_name'], '176eee2bd2c010d02dd419c453aca854195de172')

    def test_copy_raw_fields(self):
        """Test copied raw fields"""

        self._test_raw_to_enrich()
        enrich_backend = self.connectors[self.connector][2]()

        for item in self.items:
            eitem = enrich_backend.get_rich_item(item)
            for attribute in enrich_backend.RAW_FIELDS_COPY:
                if attribute in item:
                    self.assertEqual(item[attribute], eitem[attribute])
                else:
                    self.assertIsNone(eitem[attribute])

    def test_onion_study(self):
        """ Test that the onion study works correctly """

        alias = "all_onion"
        study, ocean_backend, enrich_backend = self._test_study('enrich_onion')
        study(ocean_backend, enrich_backend, alias, in_index_iss='test_github_enrich', in_index_prs="test_github_enrich",
              out_index_iss='test_github_issues_onion', out_index_prs="test_github_prs_onion")

        url = self.es_con + "/_aliases"
        response = requests.get(url, verify=False).json()
        self.assertTrue('test_github_issues_onion' in response)
        self.assertTrue('test_github_prs_onion' in response)

        time.sleep(1)

        url = self.es_con + "/test_github_issues_onion/_search?size=20"
        response = requests.get(url, verify=False).json()
        hits = response['hits']['hits']
        self.assertEqual(len(hits), 10)
        for hit in hits:
            source = hit['_source']
            self.assertIn('timeframe', source)
            self.assertIn('author_uuid', source)
            self.assertIn('author_name', source)
            self.assertIn('contributions', source)
            self.assertIn('metadata__timestamp', source)
            self.assertIn('project', source)
            self.assertIn('author_org_name', source)
            self.assertIn('cum_net_sum', source)
            self.assertIn('percent_cum_net_sum', source)
            self.assertIn('onion_role', source)
            self.assertIn('quarter', source)
            self.assertIn('metadata__enriched_on', source)
            self.assertIn('data_source', source)
            self.assertIn('grimoire_creation_date', source)

        url = self.es_con + "/test_github_prs_onion/_search?size=20"
        response = requests.get(url, verify=False).json()
        hits = response['hits']['hits']
        self.assertEqual(len(hits), 10)
        for hit in hits:
            source = hit['_source']
            self.assertIn('timeframe', source)
            self.assertIn('author_uuid', source)
            self.assertIn('author_name', source)
            self.assertIn('contributions', source)
            self.assertIn('metadata__timestamp', source)
            self.assertIn('project', source)
            self.assertIn('author_org_name', source)
            self.assertIn('cum_net_sum', source)
            self.assertIn('percent_cum_net_sum', source)
            self.assertIn('onion_role', source)
            self.assertIn('quarter', source)
            self.assertIn('metadata__enriched_on', source)
            self.assertIn('data_source', source)
            self.assertIn('grimoire_creation_date', source)


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, format='%(asctime)s %(message)s')
    logging.getLogger("urllib3").setLevel(logging.WARNING)
    logging.getLogger("requests").setLevel(logging.WARNING)
    unittest.main(warnings='ignore')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2023 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jose Javier Merchante Picazo <jjmerchante@cauldron.io>
#

from grimoire_elk.identities.identities import Identities


class GitIdentities(Identities):

    @staticmethod
    def _get_identity(git_user):
        identity = {}
        fields = git_user.split("<")
        identity['name'] = fields[0].strip()
        try:
            email = fields[1][:-1]
            identity['domain'] = email.split("@")[1]
        except IndexError:
            identity['domain'] = 'unknown'

        return identity

    @classmethod
    def anonymize_item(cls, item):
        """Remove or hash the fields that contain personal information"""

        item = item['data']

        if item['Author']:
            author = cls._get_identity(item['Author'])
            item['Author'] = "{} <xxxxxx@{}>".format(cls._hash(author['name']), author['domain'])
        if item['Commit']:
            commit = cls._get_identity(item['Commit'])
            item['Commit'] = "{} <xxxxxx@{}>".format(cls._hash(commit['name']), commit['domain'])



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2023 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jesus M. Gonzalez-Barahona <jgb@bitergia.com>
#


class Mapping:
    """Class for Elasticsearch mappings.

    This class will be subclassed by backends,
    which will provide specific mappings.
    """

    @staticmethod
    def get_elastic_mappings(es_major):
        """Get Elasticsearch mapping.

        Different versions of Elasticsearch may need specific versions
        of the mapping, thus using es_major to potentially provide
        different mappings.

        When the mapping depends on the version of Kibiter/Kibana,
        we will use the Elasticsearch version anyway, since they should
        be paired (same major version for 5 and larger, ES major version 2
        for Kibiter/Kibana major version 4).

        :param es_major: major version of Elasticsearch, as string
        :returns:        dictionary with a key, 'items', with the mapping
        """

        mapping = '{}'

        return {"items": mapping}



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2023 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Nitish Gupta <imnitish.ng@gmail.com>
#

import logging
import re
from urllib.parse import urlparse

from grimoirelab_toolkit.datetime import str_to_datetime

from .enrich import Enrich, metadata
from ..elastic_mapping import Mapping as BaseMapping


logger = logging.getLogger(__name__)


class Mapping(BaseMapping):

    @staticmethod
    def get_elastic_mappings(es_major):
        """Get Elasticsearch mapping.

        :param es_major: major version of Elasticsearch, as string
        :returns:        dictionary with a key, 'items', with the mapping
        """

        mapping = """
        {
            "properties": {
                "text_analyzed": {
                  "type": "text",
                  "fielddata": true,
                  "index": true
                }
           }
        } """

        return {"items": mapping}


class GitterEnrich(Enrich):

    mapping = Mapping

    # REGEX to extract links from HTML text
    HTML_LINK_REGEX = re.compile("href=[\"\'](.*?)[\"\']")

    def __init__(self, db_sortinghat=None, json_projects_map=None,
                 db_user='', db_password='', db_host='', db_path=None,
                 db_port=None, db_ssl=False, db_verify_ssl=True, db_tenant=None):
        super().__init__(db_sortinghat=db_sortinghat, json_projects_map=json_projects_map,
                         db_user=db_user, db_password=db_password, db_host=db_host,
                         db_port=db_port, db_path=db_path, db_ssl=db_ssl, db_verify_ssl=db_verify_ssl,
                         db_tenant=db_tenant)

    def get_field_author(self):
        return "fromUser"

    def get_sh_identity(self, item, identity_field=None):
        # email not available for gitter
        identity = {
            'username': None,
            'name': None,
            'email': None
        }

        if self.get_field_author() not in item['data']:
            return identity
        from_ = item['data'][self.get_field_author()]

        identity['username'] = from_.get('username', None)
        identity['name'] = from_.get('displayName', None)

        return identity

    def get_identities(self, item):
        """ Return the identities from an item """

        identity = self.get_sh_identity(item)
        yield identity

    def get_project_repository(self, eitem):
        tokens = eitem['origin'].rsplit("/", 1)
        return tokens[0] + " " + tokens[1]

    @metadata
    def get_rich_item(self, item):

        eitem = {}

        self.copy_raw_fields(self.RAW_FIELDS_COPY, item, eitem)

        message = item['data']

        eitem['unread'] = 1 if message['unread'] else 0
        eitem['text_analyzed'] = message['text']

        copy_fields = ["readBy", "issues", "id"]

        for f in copy_fields:
            if f in message:
                eitem[f] = message[f]
            else:
                eitem[f] = None

        eitem.update(self.get_rich_links(item['data']))

        message_timestamp = str_to_datetime(eitem['metadata__updated_on'])
        eitem['tz'] = int(message_timestamp.strftime("%H"))

        if self.sortinghat:
            eitem.update(self.get_item_sh(item))

        if self.prjs_map:
            eitem.update(self.get_item_project(eitem))

        eitem.update(self.get_grimoire_fields(item["metadata__updated_on"], "message"))

        self.add_repository_labels(eitem)
        self.add_metadata_filter_raw(eitem)
        return eitem

    def get_rich_links(self, item):

        rich_item = {}

        if item['issues']:
            self.extract_issues(item['issues'], item['html'])

        if item['mentions']:
            rich_item['mentioned'] = self.extract_mentions(item['mentions'])

        rich_item['url_hostname'] = []

        if item['urls']:
            for url in item['urls']:
                url_parsed = urlparse(url['url'])
                rich_item['url_hostname'].append('{uri.scheme}://{uri.netloc}/'.format(uri=url_parsed))

        return rich_item

    def extract_issues(self, issue_pr, html_text):
        """Enrich issues or PRs mentioned in the message"""

        links_found = self.HTML_LINK_REGEX.findall(html_text)
        for i, entity in enumerate(issue_pr):
            if 'repo' in entity.keys() and links_found:
                if links_found[i].split('/')[-2] == 'issues':
                    entity['is_issue'] = entity['repo'] + ' #' + entity['number']
                elif links_found[i].split('/')[-2] == 'pull':
                    entity['is_pull'] = entity['repo'] + ' #' + entity['number']
                else:
                    continue
                entity['url'] = links_found[i]

    def extract_mentions(self, mentioned):
        """Enrich users mentioned in the message"""

        rich_mentions = []

        for usr in mentioned:
            if 'userId' in usr.keys():
                rich_mentions.append({'mentioned_username': usr['screenName'], 'mentioned_userId': usr['userId']})

        return rich_mentions



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2023 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Alvaro del Castillo <acs@bitergia.com>
#     Valerio Cosentino <valcos@bitergia.com>
#
import logging
import json
import os
import unittest

from base import TestBaseBackend, data2es
from grimoire_elk.enriched.utils import anonymize_url, REPO_LABELS
from grimoire_elk.enriched.confluence import NO_ANCESTOR_TITLE
from grimoire_elk.raw.confluence import ConfluenceOcean
from grimoire_elk.utils import get_elastic


class TestConfluence(TestBaseBackend):
    """Test Confluence backend"""

    connector = "confluence"
    ocean_index = "test_" + connector
    enrich_index = "test_" + connector + "_enrich"

    def test_has_identites(self):
        """Test value of has_identities method"""

        enrich_backend = self.connectors[self.connector][2]()
        self.assertTrue(enrich_backend.has_identities())

    def test_items_to_raw(self):
        """Test whether JSON items are properly inserted into ES"""

        result = self._test_items_to_raw()

        self.assertEqual(result['items'], 4)
        self.assertEqual(result['raw'], 4)

        # Check enriched data
        enrich_backend = self.connectors[self.connector][2]()

        item = self.items[0]
        eitem = enrich_backend.get_rich_item(item)
        self.assertListEqual(eitem['ancestors_titles'],
                             ['Title 1', 'Title 2', 'Title 3'])
        self.assertListEqual(eitem['ancestors_links'],
                             ['/spaces/TEST/title1', '/spaces/TEST/title2', '/spaces/TEST/title3'])

        item = self.items[1]
        eitem = enrich_backend.get_rich_item(item)
        self.assertListEqual(eitem['ancestors_titles'], [NO_ANCESTOR_TITLE])
        self.assertListEqual(eitem['ancestors_links'], ['/spaces/TEST/title1'])

        item = self.items[2]
        eitem = enrich_backend.get_rich_item(item)
        self.assertListEqual(eitem['ancestors_titles'], ['Title 3'])
        self.assertListEqual(eitem['ancestors_links'], ['/spaces/TEST/title3'])

        item = self.items[3]
        eitem = enrich_backend.get_rich_item(item)
        self.assertListEqual(eitem['ancestors_titles'],
                             ['Title 1', 'Title 2', 'Title 3'])
        self.assertListEqual(eitem['ancestors_links'],
                             ['/spaces/TEST/title1', '/spaces/TEST/title2', '/spaces/TEST/title3'])

    def test_raw_to_enrich(self):
        """Test whether the raw index is properly enriched"""

        result = self._test_raw_to_enrich()
        self.assertEqual(result['raw'], 4)
        self.assertEqual(result['enrich'], 4)

        enrich_backend = self.connectors[self.connector][2]()

        item = self.items[0]
        identity = enrich_backend.get_sh_identity(item, identity_field='by')
        self.assertEqual(identity['username'], 'jdoe')
        self.assertIsNone(identity['email'])
        self.assertEqual(identity['name'], 'John Doe')

        item = self.items[1]
        identity = enrich_backend.get_sh_identity(item, identity_field='by')
        self.assertEqual(identity['username'], 'jsmith')
        self.assertIsNone(identity['email'])
        self.assertEqual(identity['name'], 'John Smith')

        item = self.items[2]
        identity = enrich_backend.get_sh_identity(item, identity_field='by')
        self.assertEqual(identity['username'], 'anonymous')
        self.assertIsNone(identity['email'])
        self.assertEqual(identity['name'], 'Anonymous')

        item = self.items[3]
        identity = enrich_backend.get_sh_identity(item, identity_field='by')
        self.assertIsNone(identity['username'])
        self.assertIsNone(identity['email'])
        self.assertEqual(identity['name'], 'Anonymous')

    def test_enrich_repo_labels(self):
        """Test whether the field REPO_LABELS is present in the enriched items"""

        self._test_raw_to_enrich()
        enrich_backend = self.connectors[self.connector][2]()

        for item in self.items:
            eitem = enrich_backend.get_rich_item(item)
            self.assertIn(REPO_LABELS, eitem)

    def test_perceval_params(self):
        """Test the extraction of perceval params from an URL"""

        url = "http://example.com --spaces=[TEST, HOME]"
        expected_params = ["http://example.com", "--spaces", "TEST", "HOME"]
        self.assertListEqual(ConfluenceOcean.get_perceval_params_from_url(url), expected_params)

    def test_raw_to_enrich_sorting_hat(self):
        """Test enrich with SortingHat"""

        result = self._test_raw_to_enrich(sortinghat=True)
        self.assertEqual(result['raw'], 4)
        self.assertEqual(result['enrich'], 4)

        enrich_backend = self.connectors[self.connector][2]()

        url = self.es_con + "/" + self.enrich_index + "/_search"
        response = enrich_backend.requests.get(url, verify=False).json()
        for hit in response['hits']['hits']:
            source = hit['_source']
            if 'author_uuid' in source:
                self.assertIn('author_domain', source)
                self.assertIn('author_gender', source)
                self.assertIn('author_gender_acc', source)
                self.assertIn('author_org_name', source)
                self.assertIn('author_bot', source)
                self.assertIn('author_multi_org_names', source)

    def test_raw_to_enrich_projects(self):
        """Test enrich with Projects"""

        result = self._test_raw_to_enrich(projects=True)
        self.assertEqual(result['raw'], 4)
        self.assertEqual(result['enrich'], 4)

    def test_copy_raw_fields(self):
        """Test copied raw fields"""

        self._test_raw_to_enrich()
        enrich_backend = self.connectors[self.connector][2]()

        for item in self.items:
            eitem = enrich_backend.get_rich_item(item)
            for attribute in enrich_backend.RAW_FIELDS_COPY:
                if attribute in item:
                    self.assertEqual(item[attribute], eitem[attribute])
                else:
                    self.assertIsNone(eitem[attribute])

    def test_refresh_identities(self):
        """Test refresh identities"""

        result = self._test_refresh_identities()
        # ... ?

    def test_raw_fix_item(self):
        """Test fix item to anonymize fields"""

        with open(os.path.join("data", self.connector + "_with_credentials.json")) as f:
            self.items = json.load(f)

        self.ocean_backend = self.connectors[self.connector][1](None)

        for item in self.items:
            # Check that the fields are not anonymized.
            self.assertNotEqual(item['origin'], anonymize_url(item['origin']))
            self.assertNotEqual(item['tag'], anonymize_url(item['tag']))
            self.assertNotEqual(item['data']['content_url'], anonymize_url(item['data']['content_url']))

            # Anonymize fields (origin, tag, data.content_url) using '_fix_item' method.
            self.ocean_backend._fix_item(item)

            # Check that the fields are anonymized.
            self.assertEqual(item['origin'], anonymize_url(item['origin']))
            self.assertEqual(item['tag'], anonymize_url(item['tag']))
            self.assertEqual(item['data']['content_url'], anonymize_url(item['data']['content_url']))

    def test_items_to_raw_anonymized_fields(self):
        """Test that the documents stored in the raw index the fields are anonymized"""

        with open(os.path.join("data", self.connector + "_with_credentials.json")) as f:
            self.items = json.load(f)

        self.ocean_backend = self.connectors[self.connector][1](None)
        elastic_ocean = get_elastic(self.es_con, self.ocean_index, True, self.ocean_backend, self.ocean_aliases)
        self.ocean_backend.set_elastic(elastic_ocean)

        # Anonymize the fields using 'anonymize_url' method.
        expected_anonymized_fields = []
        for item in self.items:
            new = {
                'origin': anonymize_url(item['origin']),
                'tag': anonymize_url(item['tag']),
                'content_url': anonymize_url(item['data']['content_url'])
            }
            expected_anonymized_fields.append(new)

        # Put items into the raw index.
        expected_items = len(self.items)
        raw_items = data2es(self.items, self.ocean_backend)
        self.assertEqual(raw_items, expected_items)

        # Fetch items from the raw index.
        raw_fields = []
        for item in self.ocean_backend.fetch():
            new = {
                'origin': item['origin'],
                'tag': item['tag'],
                'content_url': item['data']['content_url']
            }
            raw_fields.append(new)

        # Check that the fields are anonymized.
        self.assertListEqual(raw_fields, expected_anonymized_fields)


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, format='%(asctime)s %(message)s')
    logging.getLogger("urllib3").setLevel(logging.WARNING)
    logging.getLogger("requests").setLevel(logging.WARNING)
    unittest.main(warnings='ignore')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2023 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Alvaro del Castillo <acs@bitergia.com>
#     Valerio Cosentino <valcos@bitergia.com>
#     Venu Vardhan Reddy Tekula <venu@bitergia.com>
#


import logging
# import time
import unittest
from unittest.mock import MagicMock

from base import (TestBaseBackend, refresh_identities, data2es,
                  DB_SORTINGHAT, DB_HOST, FILE_PROJECTS)
from grimoire_elk.enriched.utils import REPO_LABELS
from grimoire_elk.utils import get_elastic
from grimoire_elk.elk import load_identities

from perceval.backends.core.discourse import Discourse


class TestDiscourse(TestBaseBackend):
    """Test Discourse backend"""

    connector = "discourse"
    ocean_index = "test_" + connector
    enrich_index = "test_" + connector + "_enrich"

    def test_has_identites(self):
        """Test value of has_identities method"""

        enrich_backend = self.connectors[self.connector][2]()
        self.assertTrue(enrich_backend.has_identities())

    def test_items_to_raw(self):
        """Test whether JSON items are properly inserted into ES"""

        result = self._test_items_to_raw()
        self.assertEqual(result['items'], 3)
        self.assertEqual(result['raw'], 3)

    def _test_raw_to_enrich(self, sortinghat=False, projects=False, pair_programming=False):
        """Test whether raw indexes are properly enriched"""

        # populate raw index
        perceval_backend = None
        clean = True
        self.ocean_backend = self.connectors[self.connector][1](perceval_backend)
        elastic_ocean = get_elastic(self.es_con, self.ocean_index, clean, self.ocean_backend)
        self.ocean_backend.set_elastic(elastic_ocean)
        data2es(self.items, self.ocean_backend)

        # populate enriched index
        if not sortinghat and not projects:
            self.enrich_backend = self.connectors[self.connector][2]()
        elif sortinghat and not projects:
            self.enrich_backend = self.connectors[self.connector][2](db_sortinghat=DB_SORTINGHAT,
                                                                     db_user=self.db_user,
                                                                     db_password=self.db_password,
                                                                     db_host=DB_HOST)
        elif sortinghat and projects:
            self.enrich_backend = self.connectors[self.connector][2](json_projects_map=FILE_PROJECTS,
                                                                     db_sortinghat=DB_SORTINGHAT,
                                                                     db_user=self.db_user,
                                                                     db_password=self.db_password,
                                                                     db_host=DB_HOST)
        elif not sortinghat and projects:
            self.enrich_backend = self.connectors[self.connector][2](json_projects_map=FILE_PROJECTS,
                                                                     db_user=self.db_user,
                                                                     db_password=self.db_password,
                                                                     db_host=DB_HOST)
        if pair_programming:
            self.enrich_backend.pair_programming = pair_programming

        elastic_enrich = get_elastic(self.es_con, self.enrich_index, clean, self.enrich_backend, self.enrich_aliases)
        self.enrich_backend.set_elastic(elastic_enrich)

        categories = {1: 'General', 6: 'Technical', 2: 'Ecosystem', 3: 'Staff'}
        self.enrich_backend.categories = MagicMock(return_value=categories)

        categories_tree = {1: {}, 6: {}, 2: {}, 3: {}}
        self.enrich_backend.categories_tree = MagicMock(return_value=categories_tree)

        # Load SH identities
        if sortinghat:
            load_identities(self.ocean_backend, self.enrich_backend)

        raw_count = len([item for item in self.ocean_backend.fetch()])
        enrich_count = self.enrich_backend.enrich_items(self.ocean_backend)
        # self._test_csv_mappings(sortinghat)

        return {'raw': raw_count, 'enrich': enrich_count}

    def test_raw_to_enrich(self):
        """Test whether the raw index is properly enriched"""

        result = self._test_raw_to_enrich()
        self.assertEqual(result['raw'], 3)
        self.assertEqual(result['enrich'], 35)

        enrich_backend = self.connectors[self.connector][2]()

        categories = {1: 'General', 6: 'Technical', 2: 'Ecosystem', 3: 'Staff'}
        enrich_backend.categories = MagicMock(return_value=categories)

        categories_tree = {1: {}, 6: {}, 2: {}, 3: {}}
        enrich_backend.categories_tree = MagicMock(return_value=categories_tree)

        item = self.items[0]
        eitem = enrich_backend.get_rich_item(item)
        self.assertIn('question_title', eitem)
        self.assertIn('question_title_analyzed', eitem)

    def test_enrich_repo_labels(self):
        """Test whether the field REPO_LABELS is present in the enriched items"""

        self._test_raw_to_enrich()
        enrich_backend = self.connectors[self.connector][2]()

        categories = {1: 'General', 6: 'Technical', 2: 'Ecosystem', 3: 'Staff'}
        enrich_backend.categories = MagicMock(return_value=categories)

        categories_tree = {1: {}, 6: {}, 2: {}, 3: {}}
        enrich_backend.categories_tree = MagicMock(return_value=categories_tree)

        for item in self.items:
            eitem = enrich_backend.get_rich_item(item)
            self.assertIn(REPO_LABELS, eitem)

    def test_raw_to_enrich_sorting_hat(self):
        """Test enrich with SortingHat"""

        result = self._test_raw_to_enrich(sortinghat=True)
        self.assertEqual(result['raw'], 3)
        self.assertEqual(result['enrich'], 35)

        enrich_backend = self.connectors[self.connector][2]()

        categories = {1: 'General', 6: 'Technical', 2: 'Ecosystem', 3: 'Staff'}
        enrich_backend.categories = MagicMock(return_value=categories)

        categories_tree = {1: {}, 6: {}, 2: {}, 3: {}}
        enrich_backend.categories_tree = MagicMock(return_value=categories_tree)

        url = self.es_con + "/" + self.enrich_index + "/_search"
        response = enrich_backend.requests.get(url, verify=False).json()
        for hit in response['hits']['hits']:
            source = hit['_source']
            if 'author_uuid' in source:
                self.assertIn('author_domain', source)
                self.assertIn('author_gender', source)
                self.assertIn('author_gender_acc', source)
                self.assertIn('author_org_name', source)
                self.assertIn('author_bot', source)
                self.assertIn('author_multi_org_names', source)

    def test_raw_to_enrich_projects(self):
        """Test enrich with Projects"""

        result = self._test_raw_to_enrich(projects=True)
        self.assertEqual(result['raw'], 3)
        self.assertEqual(result['enrich'], 35)

    def test_copy_raw_fields(self):
        """Test copied raw fields"""

        self._test_raw_to_enrich()
        enrich_backend = self.connectors[self.connector][2]()

        categories = {1: 'General', 6: 'Technical', 2: 'Ecosystem', 3: 'Staff'}
        enrich_backend.categories = MagicMock(return_value=categories)

        categories_tree = {1: {}, 6: {}, 2: {}, 3: {}}
        enrich_backend.categories_tree = MagicMock(return_value=categories_tree)

        for item in self.items:
            eitem = enrich_backend.get_rich_item(item)
            for attribute in enrich_backend.RAW_FIELDS_COPY:
                if attribute in item:
                    self.assertEqual(item[attribute], eitem[attribute])
                else:
                    self.assertIsNone(eitem[attribute])

    def test_refresh_identities(self):
        """Test refresh identities"""

        # populate raw index
        perceval_backend = Discourse('https://example.com', api_token='1234', api_username='user')
        clean = True
        self.ocean_backend = self.connectors[self.connector][1](perceval_backend)
        elastic_ocean = get_elastic(self.es_con, self.ocean_index, clean, self.ocean_backend)
        self.ocean_backend.set_elastic(elastic_ocean)
        data2es(self.items, self.ocean_backend)

        # populate enriched index
        self.enrich_backend = self.connectors[self.connector][2]()
        load_identities(self.ocean_backend, self.enrich_backend)
        self.enrich_backend = self.connectors[self.connector][2](db_sortinghat=DB_SORTINGHAT,
                                                                 db_user=self.db_user,
                                                                 db_password=self.db_password,
                                                                 db_host=DB_HOST)
        elastic_enrich = get_elastic(self.es_con, self.enrich_index, clean, self.enrich_backend)
        self.enrich_backend.set_elastic(elastic_enrich)
        self.enrich_backend.enrich_items(self.ocean_backend)
        total = refresh_identities(self.enrich_backend)


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, format='%(asctime)s %(message)s')
    logging.getLogger("urllib3").setLevel(logging.WARNING)
    logging.getLogger("requests").setLevel(logging.WARNING)
    unittest.main(warnings='ignore')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2023 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#

'''Bugzilla Ocean feeder'''

from .elastic import ElasticOcean
from ..elastic_mapping import Mapping as BaseMapping


class Mapping(BaseMapping):

    @staticmethod
    def get_elastic_mappings(es_major):
        """Get Elasticsearch mapping.

        :param es_major: major version of Elasticsearch, as string
        :returns:        dictionary with a key, 'items', with the mapping
        """

        mapping = '''
         {
            "dynamic":true,
                "properties": {
                    "data": {
                        "properties": {
                            "long_desc": {
                                "dynamic": false,
                                "properties": {}
                            },
                            "short_desc": {
                                "dynamic": false,
                                "properties": {
                                    "__text__": {
                                        "type": "text",
                                        "index": true
                                    }
                                }
                            },
                            "activity": {
                                "dynamic": false,
                                "properties": {}
                            }
                        }
                    }
                }
        }
        '''

        return {"items": mapping}


class BugzillaOcean(ElasticOcean):
    """Bugzilla Ocean feeder"""

    mapping = Mapping

    def _fix_item(self, item):
        # Could be used for filtering
        product = item['data']['product'][0]['__text__']
        item['product'] = product



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2023 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Valerio Cosentino <valcos@bitergia.com>
#     Miguel Ángel Fernández <mafesan@bitergia.com>
#

import logging
import time
import unittest

from base import TestBaseBackend
from grimoire_elk.enriched.utils import REPO_LABELS, anonymize_url
from grimoire_elk.raw.githubql import GitHubQLOcean
from grimoire_elk.enriched.githubql import logger


class TestGitHubQL(TestBaseBackend):
    """Test GitHubQL backend"""

    connector = "githubql"
    ocean_index = "test_" + connector
    enrich_index = "test_" + connector + "_enrich"

    def test_items_to_raw(self):
        """Test whether JSON items are properly inserted into ES"""

        result = self._test_items_to_raw()

        self.assertGreater(result['items'], 0)
        self.assertGreater(result['raw'], 0)
        self.assertEqual(result['items'], result['raw'])

    def test_raw_to_enrich(self):
        """Test whether the raw index is properly enriched"""

        result = self._test_raw_to_enrich()

        self.assertGreater(result['raw'], 0)
        self.assertGreater(result['enrich'], 0)
        self.assertEqual(result['raw'], result['enrich'])

        enrich_backend = self.connectors[self.connector][2]()

        item = self.items[0]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(item['category'], 'event')
        self.assertEqual(eitem['event_type'], 'LabeledEvent')
        self.assertEqual(eitem['created_at'], '2020-04-07T11:21:12Z')
        self.assertEqual(eitem['author_uuid'], 'ee5d85148ccdeab3efc341cb12fc70ae6b3236ae')
        self.assertEqual(eitem['author_name'], 'Unknown')
        self.assertEqual(eitem['author_user_name'], 'valeriocos')
        self.assertIsNone(eitem['author_domain'])
        self.assertEqual(eitem['repository'], 'https://github.com/valeriocos/test-issues-update')
        self.assertEqual(eitem['title'], 'Issue 2')
        self.assertEqual(eitem['label'], 'bug')
        self.assertEqual(eitem['label_description'], "Something isn't working")
        self.assertTrue(eitem['label_is_default'])
        self.assertEqual(eitem['label_created_at'], '2020-04-07T10:30:46Z')
        self.assertEqual(eitem['label_updated_at'], '2020-04-07T10:30:46Z')
        self.assertEqual(eitem['reporter_uuid'], 'ee5d85148ccdeab3efc341cb12fc70ae6b3236ae')
        self.assertEqual(eitem['reporter_name'], 'Unknown')
        self.assertEqual(eitem['reporter_user_name'], 'valeriocos')
        self.assertIsNone(eitem['reporter_domain'])

        item = self.items[1]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(item['category'], 'event')
        self.assertEqual(eitem['event_type'], 'UnlabeledEvent')
        self.assertEqual(eitem['created_at'], '2020-04-09T11:21:32Z')
        self.assertEqual(eitem['author_uuid'], 'ee5d85148ccdeab3efc341cb12fc70ae6b3236ae')
        self.assertEqual(eitem['author_name'], 'Unknown')
        self.assertEqual(eitem['author_user_name'], 'valeriocos')
        self.assertIsNone(eitem['author_domain'])
        self.assertEqual(eitem['repository'], 'https://github.com/valeriocos/test-issues-update')
        self.assertEqual(eitem['title'], 'Issue 2')
        self.assertEqual(eitem['label'], 'bug')
        self.assertEqual(eitem['label_description'], "Something isn't working")
        self.assertTrue(eitem['label_is_default'])
        self.assertEqual(eitem['label_created_at'], '2020-04-07T10:30:46Z')
        self.assertEqual(eitem['label_updated_at'], '2020-04-07T10:30:46Z')
        self.assertEqual(eitem['reporter_uuid'], 'ee5d85148ccdeab3efc341cb12fc70ae6b3236ae')
        self.assertEqual(eitem['reporter_name'], 'Unknown')
        self.assertEqual(eitem['reporter_user_name'], 'valeriocos')
        self.assertIsNone(eitem['reporter_domain'])

        item = self.items[2]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(item['category'], 'event')
        self.assertEqual(eitem['event_type'], 'AddedToProjectEvent')
        self.assertEqual(eitem['created_at'], '2020-04-07T13:22:25Z')
        self.assertEqual(eitem['author_uuid'], 'ee5d85148ccdeab3efc341cb12fc70ae6b3236ae')
        self.assertEqual(eitem['author_name'], 'Unknown')
        self.assertEqual(eitem['author_user_name'], 'valeriocos')
        self.assertIsNone(eitem['author_domain'])
        self.assertEqual(eitem['repository'], 'https://github.com/valeriocos/test-issues-update')
        self.assertEqual(eitem['title'], 'Issue 2')
        self.assertEqual(eitem['board_column'], 'backlog')
        self.assertEqual(eitem['board_name'], "my fantastic board")
        self.assertEqual(eitem['board_url'], 'https://github.com/valeriocos/test-issues-update/projects/1')
        self.assertEqual(eitem['board_created_at'], '2020-04-07T11:41:41Z')
        self.assertEqual(eitem['board_updated_at'], '2020-04-09T18:54:08Z')
        self.assertIsNone(eitem['board_closed_at'])
        self.assertEqual(eitem['board_state'], 'open')
        self.assertEqual(eitem['reporter_uuid'], 'ee5d85148ccdeab3efc341cb12fc70ae6b3236ae')
        self.assertEqual(eitem['reporter_name'], 'Unknown')
        self.assertEqual(eitem['reporter_user_name'], 'valeriocos')
        self.assertIsNone(eitem['reporter_domain'])

        item = self.items[3]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(item['category'], 'event')
        self.assertEqual(eitem['event_type'], 'MovedColumnsInProjectEvent')
        self.assertEqual(eitem['created_at'], '2020-04-07T13:22:51Z')
        self.assertEqual(eitem['author_uuid'], 'ee5d85148ccdeab3efc341cb12fc70ae6b3236ae')
        self.assertEqual(eitem['author_name'], 'Unknown')
        self.assertEqual(eitem['author_user_name'], 'valeriocos')
        self.assertIsNone(eitem['author_domain'])
        self.assertEqual(eitem['repository'], 'https://github.com/valeriocos/test-issues-update')
        self.assertEqual(eitem['title'], 'Issue 2')
        self.assertEqual(eitem['board_column'], 'analysis')
        self.assertEqual(eitem['board_previous_column'], 'backlog')
        self.assertEqual(eitem['board_name'], "my fantastic board")
        self.assertEqual(eitem['board_url'], 'https://github.com/valeriocos/test-issues-update/projects/1')
        self.assertEqual(eitem['board_created_at'], '2020-04-07T11:41:41Z')
        self.assertEqual(eitem['board_updated_at'], '2020-04-09T18:54:08Z')
        self.assertIsNone(eitem['board_closed_at'])
        self.assertEqual(eitem['board_state'], 'open')
        self.assertEqual(eitem['reporter_uuid'], 'ee5d85148ccdeab3efc341cb12fc70ae6b3236ae')
        self.assertEqual(eitem['reporter_name'], 'Unknown')
        self.assertEqual(eitem['reporter_user_name'], 'valeriocos')
        self.assertIsNone(eitem['reporter_domain'])

        item = self.items[4]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(item['category'], 'event')
        self.assertEqual(eitem['event_type'], 'RemovedFromProjectEvent')
        self.assertEqual(eitem['created_at'], '2020-04-09T18:54:08Z')
        self.assertEqual(eitem['author_uuid'], 'ee5d85148ccdeab3efc341cb12fc70ae6b3236ae')
        self.assertEqual(eitem['author_name'], 'Unknown')
        self.assertEqual(eitem['author_user_name'], 'valeriocos')
        self.assertIsNone(eitem['author_domain'])
        self.assertEqual(eitem['repository'], 'https://github.com/valeriocos/test-issues-update')
        self.assertEqual(eitem['title'], 'Issue 2')
        self.assertEqual(eitem['board_column'], 'done')
        self.assertEqual(eitem['board_name'], "my fantastic board")
        self.assertEqual(eitem['board_url'], 'https://github.com/valeriocos/test-issues-update/projects/1')
        self.assertEqual(eitem['board_created_at'], '2020-04-07T11:41:41Z')
        self.assertEqual(eitem['board_updated_at'], '2020-04-09T18:54:08Z')
        self.assertIsNone(eitem['board_closed_at'])
        self.assertEqual(eitem['board_state'], 'open')
        self.assertEqual(eitem['reporter_uuid'], 'ee5d85148ccdeab3efc341cb12fc70ae6b3236ae')
        self.assertEqual(eitem['reporter_name'], 'Unknown')
        self.assertEqual(eitem['reporter_user_name'], 'valeriocos')
        self.assertIsNone(eitem['reporter_domain'])

        item = self.items[5]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(item['category'], 'event')
        self.assertEqual(eitem['event_type'], 'CrossReferencedEvent')
        self.assertEqual(eitem['created_at'], '2020-04-07T10:36:16Z')
        self.assertEqual(eitem['author_uuid'], 'ee5d85148ccdeab3efc341cb12fc70ae6b3236ae')
        self.assertEqual(eitem['author_name'], 'Unknown')
        self.assertEqual(eitem['author_user_name'], 'valeriocos')
        self.assertIsNone(eitem['author_domain'])
        self.assertEqual(eitem['repository'], 'https://github.com/valeriocos/test-issues-update')
        self.assertEqual(eitem['title'], 'First issue')
        self.assertFalse(eitem['reference_cross_repo'])
        self.assertFalse(eitem['reference_will_close_target'])
        self.assertEqual(eitem['reference_source_type'], "Issue")
        self.assertEqual(eitem['reference_source_number'], 2)
        self.assertEqual(eitem['reference_source_repo'], 'valeriocos/test-issues-update')
        self.assertEqual(eitem['reference_source_created_at'], '2020-04-07T10:35:56Z')
        self.assertEqual(eitem['reference_source_updated_at'], '2020-04-07T11:21:32Z')
        self.assertIsNone(eitem['reference_source_closed_at'])
        self.assertFalse(eitem['reference_source_closed'])
        self.assertIsNone(eitem['reference_source_merged'])
        self.assertEqual(eitem['reporter_uuid'], 'ee5d85148ccdeab3efc341cb12fc70ae6b3236ae')
        self.assertEqual(eitem['reporter_name'], 'Unknown')
        self.assertEqual(eitem['reporter_user_name'], 'valeriocos')
        self.assertIsNone(eitem['reporter_domain'])

        item = self.items[6]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(item['category'], 'event')
        self.assertEqual(eitem['event_type'], 'ClosedEvent')
        self.assertEqual(eitem['created_at'], '2020-04-12T07:51:05Z')
        self.assertEqual(eitem['author_uuid'], 'ee5d85148ccdeab3efc341cb12fc70ae6b3236ae')
        self.assertEqual(eitem['author_name'], 'Unknown')
        self.assertEqual(eitem['author_user_name'], 'valeriocos')
        self.assertIsNone(eitem['author_domain'])
        self.assertEqual(eitem['repository'], 'https://github.com/valeriocos/test-issues-update')
        self.assertEqual(eitem['title'], 'First issue')
        self.assertEqual(eitem['closer_type'], "PullRequest")
        self.assertEqual(eitem['closer_number'], 3)
        self.assertEqual(eitem['closer_repo'], "valeriocos/test-issues-update")
        self.assertEqual(eitem['closer_created_at'], '2020-04-12T07:50:25Z')
        self.assertEqual(eitem['closer_updated_at'], '2020-04-12T07:51:05Z')
        self.assertEqual(eitem['closer_closed_at'], '2020-04-12T07:51:05Z')
        self.assertTrue(eitem['closer_closed'])
        self.assertTrue(eitem['closer_merged'])
        self.assertEqual(eitem['reporter_uuid'], 'ee5d85148ccdeab3efc341cb12fc70ae6b3236ae')
        self.assertEqual(eitem['reporter_name'], 'Unknown')
        self.assertEqual(eitem['reporter_user_name'], 'valeriocos')
        self.assertIsNone(eitem['reporter_domain'])
        self.assertEqual(eitem['submitter_uuid'], 'ee5d85148ccdeab3efc341cb12fc70ae6b3236ae')
        self.assertEqual(eitem['submitter_name'], 'Unknown')
        self.assertEqual(eitem['submitter_user_name'], 'valeriocos')
        self.assertIsNone(eitem['submitter_domain'])
        self.assertEqual(eitem['closer_pull_submitter'], 'valeriocos')

        item = self.items[13]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(item['category'], 'event')
        self.assertEqual(eitem['event_type'], 'MergedEvent')
        self.assertEqual(eitem['created_at'], '2020-10-23T18:45:14Z')
        self.assertEqual(eitem['author_uuid'], '09c94d1fdc9523c9db2244d7ae4d1ab9603cd683')
        self.assertEqual(eitem['author_name'], 'Unknown')
        self.assertEqual(eitem['author_user_name'], 'zhquan')
        self.assertIsNone(eitem['author_domain'])
        self.assertEqual(eitem['repository'], 'https://github.com/zhquan/test-merged-event')
        self.assertEqual(eitem['title'], 'Initial directory setup')
        self.assertTrue(eitem['merge_closed'])
        self.assertEqual(eitem['merge_closed_at'], '2020-10-23T18:45:14Z')
        self.assertEqual(eitem['merge_created_at'], '2020-10-23T18:45:06Z')
        self.assertTrue(eitem['merge_merged'])
        self.assertEqual(eitem['merge_merged_at'], '2020-10-23T18:45:14Z')
        self.assertEqual(eitem['merge_updated_at'], '2020-10-23T18:45:14Z')
        self.assertEqual(eitem['merge_url'], 'https://github.com/zhquan/test-merged-event/pull/3')
        self.assertNotIn('merge_approved', eitem)

        item = self.items[15]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(item['category'], 'event')
        self.assertEqual(eitem['event_type'], 'PullRequestReview')
        self.assertEqual(eitem['created_at'], '2021-08-06T10:33:30Z')
        self.assertEqual(eitem['author_uuid'], '09c94d1fdc9523c9db2244d7ae4d1ab9603cd683')
        self.assertEqual(eitem['author_name'], 'Unknown')
        self.assertEqual(eitem['author_user_name'], 'zhquan')
        self.assertIsNone(eitem['author_domain'])
        self.assertEqual(eitem['repository'], 'https://github.com/zhquan/test-merged-event')
        self.assertEqual(eitem['title'], 'commit')
        self.assertTrue(eitem['merge_closed'])
        self.assertEqual(eitem['merge_closed_at'], '2021-08-06T10:34:32Z')
        self.assertEqual(eitem['merge_created_at'], '2021-08-03T10:52:58Z')
        self.assertTrue(eitem['merge_merged'])
        self.assertEqual(eitem['merge_merged_at'], '2021-08-06T10:34:31Z')
        self.assertEqual(eitem['merge_updated_at'], '2021-08-06T10:34:32Z')
        self.assertEqual(eitem['merge_url'], 'https://github.com/zhquan/test-merged-event/pull/442')
        self.assertEqual(eitem['merge_state'], 'APPROVED')
        self.assertEqual(eitem['merge_approved'], 1)

    def test_enrich_repo_labels(self):
        """Test whether the field REPO_LABELS is present in the enriched items"""

        self._test_raw_to_enrich()
        enrich_backend = self.connectors[self.connector][2]()

        for item in self.items:
            eitem = enrich_backend.get_rich_item(item)
            self.assertIn(REPO_LABELS, eitem)

    def test_raw_to_enrich_sorting_hat(self):
        """Test enrich with SortingHat"""

        result = self._test_raw_to_enrich(sortinghat=True)
        self.assertGreater(result['raw'], 0)
        self.assertGreater(result['enrich'], 0)
        self.assertEqual(result['raw'], result['enrich'])

        enrich_backend = self.connectors[self.connector][2]()

        url = self.es_con + "/" + self.enrich_index + "/_search"
        response = enrich_backend.requests.get(url, verify=False).json()
        for hit in response['hits']['hits']:
            source = hit['_source']
            if 'author_uuid' in source:
                self.assertIn('author_domain', source)
                self.assertIn('author_gender', source)
                self.assertIn('author_gender_acc', source)
                self.assertIn('author_org_name', source)
                self.assertIn('author_bot', source)
                self.assertIn('author_multi_org_names', source)
            if 'reporter_uuid' in source:
                self.assertIn('reporter_domain', source)
                self.assertIn('reporter_gender', source)
                self.assertIn('reporter_gender_acc', source)
                self.assertIn('reporter_org_name', source)
                self.assertIn('reporter_bot', source)
                self.assertIn('reporter_multi_org_names', source)
            if 'submitter_uuid' in source:
                self.assertIn('submitter_domain', source)
                self.assertIn('submitter_gender', source)
                self.assertIn('submitter_gender_acc', source)
                self.assertIn('submitter_org_name', source)
                self.assertIn('submitter_bot', source)
                self.assertIn('submitter_multi_org_names', source)

    def test_raw_to_enrich_projects(self):
        """Test enrich with Projects"""

        result = self._test_raw_to_enrich(projects=True)
        self.assertGreater(result['raw'], 0)
        self.assertGreater(result['enrich'], 0)
        self.assertEqual(result['raw'], result['enrich'])

        enrich_backend = self.connectors[self.connector][2]()

        url = self.es_con + "/" + self.enrich_index + "/_search"
        response = enrich_backend.requests.get(url, verify=False).json()
        for hit in response['hits']['hits']:
            source = hit['_source']
            self.assertEqual(source['project'], 'Main')
            self.assertEqual(source['project_1'], 'Main')

    def test_refresh_identities(self):
        """Test refresh identities"""

        result = self._test_refresh_identities()
        # ... ?

    def test_perceval_params(self):
        """Test the extraction of perceval params from an URL"""

        url = "https://github.com/chaoss/grimoirelab-perceval"
        expected_params = [
            'chaoss', 'grimoirelab-perceval'
        ]
        self.assertListEqual(GitHubQLOcean.get_perceval_params_from_url(url), expected_params)

    def test_copy_raw_fields(self):
        """Test copied raw fields"""

        self._test_raw_to_enrich()
        enrich_backend = self.connectors[self.connector][2]()

        for item in self.items:
            eitem = enrich_backend.get_rich_item(item)
            for attribute in enrich_backend.RAW_FIELDS_COPY:
                if attribute in item:
                    self.assertEqual(item[attribute], eitem[attribute])
                else:
                    self.assertIsNone(eitem[attribute])

    def test_duration_analysis(self):
        """Test that the geolocation study works correctly"""

        study, ocean_backend, enrich_backend = self._test_study('enrich_duration_analysis')

        with self.assertLogs(logger, level='INFO') as cm:

            if study.__name__ == "enrich_duration_analysis":
                study(ocean_backend, enrich_backend,
                      start_event_type="UnlabeledEvent", target_attr="label",
                      fltr_attr="label", fltr_event_types=["LabeledEvent"])

            self.assertEqual(cm.output[0], 'INFO:grimoire_elk.enriched.githubql:[githubql] Duration analysis '
                                           'starting study %s/test_githubql_enrich'
                             % anonymize_url(self.es_con))
            self.assertEqual(cm.output[-1], 'INFO:grimoire_elk.enriched.githubql:[githubql] Duration analysis '
                                            'ending study %s/test_githubql_enrich'
                             % anonymize_url(self.es_con))

        time.sleep(5)  # HACK: Wait until github enrich index has been written
        items = [item for item in enrich_backend.fetch() if item['event_type'] == 'UnlabeledEvent']
        self.assertEqual(len(items), 1)
        for item in items:
            self.assertEqual(item['previous_event_uuid'], 'f371d54454d297f86f08ab52a440ae5f9e4afeb1')
            self.assertEqual(item['duration_from_previous_event'], 2.0)

    def test_reference_analysis(self):
        """Test that the cross reference study works correctly"""

        study, ocean_backend, enrich_backend = self._test_study('enrich_reference_analysis')

        with self.assertLogs(logger, level='INFO') as cm:

            if study.__name__ == "enrich_reference_analysis":
                study(ocean_backend, enrich_backend)

            self.assertEqual(cm.output[0], 'INFO:grimoire_elk.enriched.githubql:[githubql] Cross reference analysis '
                                           'starting study %s/test_githubql_enrich'
                             % anonymize_url(self.es_con))
            self.assertEqual(cm.output[-1], 'INFO:grimoire_elk.enriched.githubql:[githubql] Cross reference analysis '
                                            'ending study %s/test_githubql_enrich'
                             % anonymize_url(self.es_con))

        time.sleep(5)  # HACK: Wait until github enrich index has been written
        referenced_items = []
        for item in enrich_backend.fetch():
            if ('referenced_by_issues' or 'referenced_by_prs') in item.keys():
                referenced_items.append(item)

        self.assertEqual(len(referenced_items), 7)

        for item in referenced_items:
            self.assertIn('referenced_by_issues', item)
            self.assertIn('referenced_by_prs', item)
            self.assertIn('referenced_by_merged_prs', item)
            self.assertIn('referenced_by_external_issues', item)
            self.assertIn('referenced_by_external_prs', item)
            self.assertIn('referenced_by_external_merged_prs', item)

            ref_issues = item['referenced_by_issues']
            self.assertEqual(len(ref_issues), 1)

            ref = ref_issues[0]
            self.assertEqual(ref, 'https://github.com/valeriocos/test-issues-update/issues/2')

            ref_prs = item['referenced_by_prs']
            self.assertEqual(len(ref_prs), 1)

            ref = ref_prs[0]
            self.assertEqual(ref, 'https://github.com/valeriocos/test-issues-update/pull/3')

            ref_merged_prs = item['referenced_by_merged_prs']
            self.assertEqual(len(ref_merged_prs), 1)

            ref = ref_merged_prs[0]
            self.assertEqual(ref, 'https://github.com/valeriocos/test-issues-update/pull/3')

            ref_ext_issues = item['referenced_by_external_issues']
            self.assertEqual(len(ref_ext_issues), 0)

            ref_ext_prs = item['referenced_by_external_prs']
            self.assertEqual(len(ref_ext_prs), 0)

            ref_ext_merged_prs = item['referenced_by_external_merged_prs']
            self.assertEqual(len(ref_ext_merged_prs), 0)


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, format='%(asctime)s %(message)s')
    logging.getLogger("urllib3").setLevel(logging.WARNING)
    logging.getLogger("requests").setLevel(logging.WARNING)
    unittest.main(warnings='ignore')
